# -*- coding: utf-8 -*-
import enum

from qtpy.QtCore import Qt

DEFAULT_SCENE_WIDTH = 1000
DEFAULT_SCENE_HEIGHT = 1000


#class Lock(enum.IntFlag):
class Lock(object):
    """
    For bitwise operations, do not use Enum class
    """
    NODES = 1
    LINKS = 2
    LABELS = 4


#class DataRole(enum.IntEnum):
class DataRole(object):
    InitStatusRole = Qt.UserRole + 1



from qtpy.QtCore import Qt
from qtpy.QtCore import QObject
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QGraphicsRectItem

from nezzle.graphics.screen import GraphicsScene
from nezzle.graphics.screen import GraphicsView
from nezzle.constants import DEFAULT_SCENE_WIDTH, DEFAULT_SCENE_HEIGHT


class SingleViewManager(QObject):

    def __init__(self, mw):
        super().__init__()
        self.mw = mw

        self._tab_widget = self.mw.ui_mainTabWidget

        self._default_scene = GraphicsScene()
        self._default_scene.setSceneRect(0, 0, DEFAULT_SCENE_WIDTH, DEFAULT_SCENE_HEIGHT)

        self._view = GraphicsView(mw)
        self._view.setScene(self._default_scene)

        self._is_empty = True
        self._tab_widget.addTab(self.view, "<Not selected>")
        self._tab_widget.setTabsClosable(False)

    @property
    def is_empty(self):
        return self._is_empty

    @property
    def tab_widget(self):
        return self._tab_widget

    @property
    def view(self):
        return self._view

    @property
    def current_scene(self):
        return self.view.scene()

    def clear(self):
        self._default_scene.clear()
        self.view.setScene(self._default_scene)
        self.set_current_view_text("<Not selected>")
        self._is_empty = True

    def set_current_view_scene(self, scene, text):
        if not scene:
            self.clear()
            return

        self.view.setScene(scene)
        self.set_current_view_text(text)
        self._is_empty = False

        """
        To see the area of scene
        
        rect_item = QGraphicsRectItem(-scene.width()/2, -scene.height()/2,
                                       scene.width(), scene.height())
        rect_item.setPen(Qt.red)
        rect_item.setPos(scene.width()/2, scene.height()/2)
        scene.addItem(rect_item)
        """

    def set_current_view_text(self, text):
        index = self.tab_widget.currentIndex()
        self.tab_widget.setTabText(index, text)





import math
import random

from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from qtpy.QtGui import QColor
from qtpy.QtGui import QFont

import networkx as nx

import nezzle

from nezzle.graphics.nodes.nodefactory import NodeClassFactory
from nezzle.graphics.links.linkfactory import LinkClassFactory
from nezzle.graphics.labels.labelfactory import LabelClassFactory
from nezzle.graphics.arrows.arrowclassfactory import ArrowClassFactory

from nezzle.graphics.links.baselink import BaseLink
from nezzle.graphics.labels.textlabel import TextLabel
from nezzle.graphics.nodes.basenode import BaseNode

from nezzle.graphics import SelfloopLink
from nezzle.graphics.baseitem import MappableItem
from nezzle.graphics.screen import GraphicsScene
from nezzle.utils.math import rotate, dist, internal_division


class Network(MappableItem):
    def __init__(self, iden, name=None):
        self._item = None
        super().__init__(iden, name=name)
        self._nodes = {}
        self._links = {}
        self._labels = {}
        self._scene = GraphicsScene()
        self._nxdg = nx.DiGraph(name=name)
        self.nxdg.labels = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return '%s (%s) object at %s' % (__class__, self.name, id(self))

    @property
    def item(self):
        return self._item

    @item.setter
    def item(self, obj):
        self._item = obj

    @property
    def nodes(self):
        return self._nodes

    @property
    def links(self):
        return self._links

    @property
    def labels(self):
        return self._labels

    @property
    def scene(self):
        return self._scene

    @property
    def nxdg(self):
        return self._nxdg

    def _trigger_set_name(self, key, value):
        self._name = value
        if self._item:
            self._item.setText(self._name)

        return value

    def add_node(self, node):
        node.setZValue(0)
        self.nodes[node.iden] = node
        self.scene.addItem(node)
        self.nxdg.add_node(node.iden)
        self.nxdg.nodes[node.iden]['GRAPHICS'] = node

    def remove_node(self, obj):
        if isinstance(obj, BaseNode):
            iden = obj.iden
        elif isinstance(obj, str):
            iden = obj

        node = self.nodes[iden]
        while node.links:
            link = node.links.pop()
            self.remove_link(link)

        self.scene.removeItem(node)
        del self.nodes[iden]

    def replace_node(self, old_node, new_node):
        while old_node.links:
            link = old_node.links.pop()  # Remove all links from old node,
            if id(link.source) == id(old_node):
                link.source = new_node
                new_node.add_link(link)
            elif id(link.target) == id(old_node):
                link.target = new_node
                new_node.add_link(link)
            else:
                raise RuntimeError("[SYSTEM] this point should not be reached!")
        # end of while

        # Add the new node.
        self.add_node(new_node)

        # Re-organize children.
        for child in old_node.childItems():
            child.setParentItem(new_node)

        # Remove the old node from the scene.
        self.scene.removeItem(old_node)

    def replace_link(self, old_link, new_link):
        if isinstance(old_link, SelfloopLink):
            self.nxdg.remove_edge(old_link.node.iden, old_link.node.iden)
            self.nxdg.add_edge(new_link.node.iden, new_link.node.iden)

            old_link.node.remove_link(old_link)
            old_link.node.add_link(new_link)
        else:
            self.nxdg.remove_edge(old_link.source.iden, old_link.target.iden)
            self.nxdg.add_edge(new_link.source.iden, new_link.target.iden)

            old_link.source.remove_link(old_link)
            old_link.target.remove_link(old_link)

            old_link.source.add_link(new_link)
            old_link.target.add_link(new_link)

        del self.links[old_link.iden]
        self.links[new_link.iden] = new_link

        old_link.setZValue(-1)
        new_link.setZValue(-1)
        self.scene.removeItem(old_link)
        self.scene.addItem(new_link)
        new_link.update()

    def add_link(self, link):
        """Add a links object.

        Args:
            link : nezzle.graphics.BaseLink
                Link object derived from nezzle.graphics.BaseLink.

        Returns:
            None
        """

        link.setZValue(-1)
        self.links[link.iden] = link
        self.scene.addItem(link)

        if isinstance(link, SelfloopLink):
            iden_src = iden_tgt = link.node.iden
        else:
            iden_src = link.source.iden
            iden_tgt = link.target.iden

        self.nxdg.add_edge(iden_src, iden_tgt)
        self.nxdg[iden_src][iden_tgt]['GRAPHICS'] = link

    def remove_link(self, obj):
        if isinstance(obj, BaseLink):
            iden = obj.iden
        elif isinstance(obj, str):
            iden = obj

        link = self.links[iden]
        self.scene.removeItem(link)
        if isinstance(obj, SelfloopLink):
            self.nxdg.remove_edge(link.node.iden, link.node.iden)
            link.node.remove_link(link)
        else:
            self.nxdg.remove_edge(link.source.iden, link.target.iden)
            link.source.remove_link(link)
            link.target.remove_link(link)

        del self.links[iden]

    def add_label(self, label):
        self.labels[label.iden] = label
        self.nxdg.labels[label.iden] = {}
        self.nxdg.labels[label.iden]['GRAPHICS'] = label

    def remove_label(self, obj):
        if isinstance(obj, TextLabel):
            iden = obj.iden
        elif isinstance(obj, str):
            iden = obj

        self.scene.removeItem(self.labels[iden])
        del self.labels[iden]

    def copy(self):
        return self.from_dict(self.to_dict())

    def to_dict(self):
        dict_net = {}

        # TODO: Use a global variable for setting the version
        dict_net["NEZZLE_VERSION"] = tuple(nezzle.__version__.split('.'))
        dict_net["NAME"] = self.name

        bg_color = self.scene.backgroundBrush().color()
        dict_net["BACKGROUND_COLOR"] = bg_color.name(QColor.HexArgb)

        dict_net["NODES"] = []
        dict_net["LINKS"] = []
        dict_net["LABELS"] = []

        for iden, node in self.nodes.items():
            dict_node = node.to_dict()
            dict_node = {key: val for key, val in dict_node.items() if not key.startswith("_")}
            dict_net["NODES"].append(dict_node)

        for iden, link in self.links.items():
            dict_link = link.to_dict()
            dict_link = {key: val for key, val in dict_link.items() if not key.startswith("_")}
            dict_net["LINKS"].append(dict_link)

        for iden, label in self.labels.items():
            dict_label = label.to_dict()
            dict_label = {key: val for key, val in dict_label.items() if not key.startswith("_")}
            dict_net["LABELS"].append(dict_label)

        dict_net.update(self._attr)
        return dict_net

    @classmethod
    def from_dict(cls, dict_net):
        """
        Adding objects should be in the following order:

            Link -> Node -> Label

        This order prevents an abnormal visualization.
        """
        # TODO: Using Z-value for maintaining the order.

        net = cls(dict_net['ID'])
        net.name = dict_net['NAME']

        bg_color_rgb = dict_net["BACKGROUND_COLOR"]
        bg_color = QColor(bg_color_rgb)
        net.scene.setBackgroundBrush(bg_color)

        list_nodes = []
        dict_graphics = {}

        for dict_node in dict_net["NODES"]:
            item_type = dict_node.pop('ITEM_TYPE')
            NodeClass = NodeClassFactory.create(item_type)
            node = NodeClass.from_dict(dict_node)
            dict_graphics[node.iden] = node
            list_nodes.append(node)

        for dict_link in dict_net["LINKS"]:
            item_type = dict_link.pop("ITEM_TYPE")
            if 'SELFLOOP' in item_type:
                iden_node = dict_link.pop("ID_NODE")
                node = dict_graphics[iden_node]
                LinkClass = LinkClassFactory.create(item_type)
                link = LinkClass.from_dict(dict_link, node)
            else:
                iden_src = dict_link.pop("ID_SOURCE")
                iden_tgt = dict_link.pop("ID_TARGET")
                src = dict_graphics[iden_src]
                tgt = dict_graphics[iden_tgt]
                LinkClass = LinkClassFactory.create(item_type)
                link = LinkClass.from_dict(dict_link, src, tgt)

            dict_graphics[link.iden] = link
            net.add_link(link)
            link.update()

        for dict_label in dict_net["LABELS"]:
            item_type = dict_label.pop("ITEM_TYPE")
            iden_parent = dict_label.pop("ID_PARENT")
            parent = dict_graphics[iden_parent]
            LabelClass = LabelClassFactory.create(item_type)
            label = LabelClass.from_dict(dict_label, parent)
            net.add_label(label)

        for node in list_nodes:
            net.add_node(node)

        return net


def from_adj_to_net(A, i2n, name='network', msc=None, nodes=None):
    """Create a network using adjacency matrix.

    Parameters
    ----------
    A : numpy.ndarray
        Adjacency matrix.
    i2n : dict
        Dictionary for mapping index to name.
        Node IDs and labels are created from this dictionary.
    name : str, optional
        Network name, also used for network identity.
    msc : dict, optional
        Dictionary for mapping signs to links classes.
        If it is not given, a positive value in the adjacency
        is used for creating arrow links, and a negative value
        for creating hammer links.
    nodes : dict, optional
        Dictionary of nodes objects (nezzle.graphics.BaseNode).
        The new nodes are created by copying these nodes.
        This dictionary should be given with i2n.


    Returns
    -------
    net : nezzle.graphics.Network
        Network object.
    """

    NodeClass = NodeClassFactory.create("ELLIPSE_NODE")
    LinkClass = LinkClassFactory.create("CURVED_LINK")
    SelfloopClass = LinkClassFactory.create("SELFLOOP_LINK")
    LabelClass = LabelClassFactory.create("TEXT_LABEL")

    if not nodes:
        nodes = {}
        range_pos = (0, 500)
        width = 50
        height = 35
        for iden in i2n.values():
            x = random.uniform(*range_pos)
            y = random.uniform(*range_pos)
            node = NodeClass(iden, pos=QPointF(x, y),
                                    width=width, height=height)

            node['BORDER_COLOR'] = Qt.darkGray

            nodes[iden] = node

    if not msc:
        msc = {}
        msc['+'] = ArrowClassFactory.create("TRIANGLE")
        msc['-'] = ArrowClassFactory.create("HAMMER")


    ir, ic = A.nonzero()
    nodes_cache = {}
    net_new = Network(name)
    # net_new.scene.setBackgroundBrush(QColor(0, 0, 0, 0))
    net_new.scene.setBackgroundBrush(Qt.white) #transparent)
    for i in range(ir.size):
        itgt, isrc = ir[i], ic[i]
        tgt, src = i2n[itgt], i2n[isrc]

        sign = '+' if A[itgt, isrc] > 0 else '-'
        ArrowClass = msc[sign]
        head = ArrowClass()

        iden = "%s%s%s" % (src, sign, tgt)

        if src == tgt:
            if src not in nodes_cache:
                node = nodes[src].copy()
                nodes_cache[src] = node
            else:
                node = nodes_cache[src]

            link = SelfloopClass(iden=iden,
                                 node=node,
                                 head=head)

            node.add_link(link)
            net_new.add_link(link)
        else:  # Two-nodes links
            if src not in nodes_cache:
                node_src = nodes[src].copy()
                nodes_cache[src] = node_src
            else:
                node_src = nodes_cache[src]

            if tgt not in nodes_cache:
                node_tgt = nodes[tgt].copy()
                nodes_cache[tgt] = node_tgt
            else:
                node_tgt = nodes_cache[tgt]

            link = LinkClass(iden=iden,
                             source=node_src,
                             target=node_tgt,
                             head=head)

            node_src.add_link(link)
            node_tgt.add_link(link)
            net_new.add_link(link)

            # end of for

    # Add labels
    font = QFont()
    font.setFamily("Tahoma")
    font.setPointSize(10)

    for str_name, node in nodes_cache.items():
        net_new.add_node(node)
        label = LabelClass(node, str_name)
        label.font = font
        rect = label.boundingRect()
        label.setPos(-rect.width() / 2, -rect.height() / 2)
        net_new.add_label(label)

    for src, tgt, attr in net_new.nxdg.edges(data=True):
        if net_new.nxdg.has_edge(tgt, src):
            if src == tgt:  # Skip selfloops
                continue

            link = attr['GRAPHICS']
            mid = internal_division(link.pos_src, link.pos_tgt, 0.5, 0.5)
            d = dist(link.pos_src, mid) / math.cos(math.pi / 4)
            cp = rotate(link.pos_src, mid, -30, d)
            link.ctrl_point.setPos(cp)

    return net_new


from qtpy.QtCore import QRectF
from qtpy.QtWidgets import QGraphicsItem

from nezzle.graphics import PainterOptionItem


class BaseNode(PainterOptionItem):

    ITEM_TYPE = 'NODE'

    def __init__(self, iden, width, height, *args, **kwargs):
        super().__init__(iden, *args, **kwargs)
        self._iden = iden
        self._brect = QRectF()
        self._links = []

        self._attr.set_trigger('WIDTH', self._trigger_set_width)
        self._attr.set_trigger('HEIGHT', self._trigger_set_height)

        self.width = width
        self.height = height

        # if pos:
        # #    self.setPos(pos)
        #     self._attr["POS_X"] = pos.x()
        #     self._attr["POS_Y"] = pos.y()

        self.setFlags(QGraphicsItem.ItemIsSelectable
                      | QGraphicsItem.ItemIsMovable
                      | QGraphicsItem.ItemIsFocusable
                      | QGraphicsItem.ItemSendsGeometryChanges)

    # end of def __init__

    def __str__(self):
        return 'Node(%s)'%(self.name)

    @property
    def links(self):
        return self._links

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, val):
        self._attr['WIDTH'] = val

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, val):
        self._attr['HEIGHT'] = val

    def is_movable(self):
        return True

    def gather_children(self):
        children = super().gather_children()
        for link in self.links:
            for child in link.childItems():
                if child not in children:
                    children.append(child)

        return children

    def boundingRect(self):
        rect = QRectF(self._brect)
        if self._pen:
            wp = self._pen.width()
            rect.adjust(-wp, -wp, +wp, +wp)

        return rect

    def update(self, *args, **kwargs):
        for link in self._links:
            link.update()
        self._invalidate()
        return super().update(*args, **kwargs)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionHasChanged:
            self.update()
        elif change == QGraphicsItem.ItemSelectedHasChanged:
            self.update()

        return super().itemChange(change, value)

    def has_link(self, link):
        return link in self._links

    def add_link(self, link):
        if self.has_link(link):
            return
        self._links.append(link)

    def remove_link(self, link):
        if link in self._links:
            self._links.remove(link)

    def copy(self):
        return self.from_dict(self.to_dict())

    def _trigger_set_width(self, key, value):
        self._width = value
        self._brect.setX(-value / 2)
        self._brect.setWidth(value)
        self.update()
        return value

    def _trigger_set_height(self, key, value):
        self._height = value
        self._brect.setY(-value / 2)
        self._brect.setHeight(value)
        self.update()
        return value

    @classmethod
    def from_dict(cls, attr):
        iden = attr.pop('ID')
        width = attr.pop('WIDTH')
        height = attr.pop('HEIGHT')

        obj = cls(iden=iden, width=width, height=height)
        obj._attr.update(attr)

        return obj



"""
Utility functions for mathematical calculation.
"""

import numpy as np

from qtpy.QtCore import QLineF
from qtpy.QtCore import QPointF
from qtpy.QtGui import QTransform
from qtpy.QtGui import QVector2D


def dot(v1, v2):
    """Calculate the dot product of two vectors.

    It is equal to v1.x()*v2.x() + v1.y()*v2.y().

    This function is a simple wrapper of
    QPointF.dotProduct(v1, v2).

    Args:
        v1 : QPointF
            First vector from the origin.
        v2 : QPointF
            Second vector from the origin.

    Returns:
        dp : float
            Dot product of the two vectors.

    Examples:
        >>> v1 = QPointF(1, 2)
        >>> v2 = QPointF(2, 4)
        >>> dot(v1, v2)
        10.0

        >>> v1 = QPointF(1, 1)
        >>> v2 = QPointF(10, 10)
        >>> dot(v1, v2)
        20.0
    """
    return QPointF.dotProduct(v1, v2)


def length(v):
    """Calculate the length of a vector.

    Args:
        v : QPointF
            A vector from the origin.

    Returns:
        len : float
            Length of the vector, v.

    Examples:
        >>> v1 = QPointF(1, 0)
        >>> length(v1)
        1.0

        >>> v1 = QPointF(1, 1)
        >>> length(v1)
        1.4142135623730951

        >>> import math
        >>> length(v1) == np.sqrt(2)
        True
    """
    return np.sqrt(v.x()**2 + v.y()**2)


def dist(p1, p2):
    """Calculate the distance between two points.

    Args:
        p1 : QPointF
            First point
        p2 : QPointF
            Second point

    Returns:
        distance : float
            Distance between the two points.

    Examples:
        >>> a = QPointF(0, 10)
        >>> b = QPointF(10, 0)
        >>> dist(a, b)  # square root of 2
        14.142135623730951

    """
    diff = p1 - p2
    return np.sqrt(diff.x()**2 + diff.y()**2)


def angle(v1, v2, radian=False):
    """Calculate the angle in degrees between two vectors.

    Args:
        v1 : QPointF
            First vector from the origin.
        v2 : QPointF
            Second vector from the origin.

    Returns:
        angle : float
            Angle between the two vectors, v1 and v2.

    Examples:
        >>> v1 = QPointF(10, 0)
        >>> v2 = QPointF(10, 10)
        >>> angle(v1, v2)
        45.00000000000001

        >>> v3 = QPointF(-10, 0)
        >>> angle(v1, v3)
        180.0

        >>> angle(v2, v3)
        135.0

        >>> v1 = QPointF()
    """

    len_v1 = length(v1)
    len_v2 = length(v2)

    theta = np.arccos(dot(v1, v2) / (len_v1 * len_v2))
    if radian:
        return theta  # in radian

    return np.degrees(theta)  # in degree

# def angle(v1, v2, radian=False):
#     """Get the angle between two vectors or points.
#     """
#     radian = np.pi / 180 if radian else 1
#
#     return QLineF(v1.x(), v1.y(), v2.x(), v2.y()).angle() * radian

def normal_vector(v):
    """Get the counter clock-wise normal vector for a given vector.
    """
    v = QVector2D(v.x(), v.y()).normalized()
    return QVector2D(-v.y(), v.x())


def internal_division(p1, p2, r1, r2):
    """Find the point internally dividing two points.

    ip = r2/(r1+r2)*p1 + r1/(r1+r2)*b

    (p1)---<r1>---(ip)---<r2>---(p2)


    Args:
        p1 : QPointF
            First point.
        p2 : QPointF
            Second point.
        r1 : float
            Ratio from the p1.
        r2 : float
            Ratio from the p2.

    Returns:
        ip : QPointF
            Internal point of the two points.

    Examples:
        >>> a = QPointF(0, 0)
        >>> b = QPointF(2, 0)
        >>> internal_division(a, b, 1, 1)
        PyQt5.QtCore.QPointF(1.0, 0.0)

        >>> c = QPointF(1, 1)
        >>> internal_division(a, c, 0.3, 0.7)
        PyQt5.QtCore.QPointF(0.3, 0.3)

        >>> internal_division(a, b, 2, 8)
        PyQt5.QtCore.QPointF(0.4, 0.0)
    """
    sum_r = r1 + r2


    return (r2/sum_r*p1) + (r1/sum_r*p2)


def solve_cubic(a, b, c, d):
    """Solve a cubic equation: a*x^3 + b*x^2+ c*x + d

    This is a Python version of the cubic equation solver [#F1]_.

    Args:
        a : float
            Coefficient of x^3.
        b : float
            Coefficient of x^2.
        c : float
            Coefficient of x.
        d : float
            Constant term.

    Returns:
        x : list
            Solutions of the cubic equation.
            The number of solutions is three at most.

    Examples:
        >>> solve_cubic(1, 0, 0, -1)
        [1.0]

        >>> solve_cubic(1, 0, 0, 1)
        [-1.0]

        >>> solve_cubic(1, -1, -4, 4)
        [1.9999999999999993, 1.0, -1.9999999999999998]

        >>> solve_cubic(3, -1, 1, 2)
        [-0.6666666666666663]

        >>> solve_cubic(1, -4, 3, 0)
        [3.0, 0.9999999999999997, 0.0]

    References:
         .. [#F1] Glassner, "Graphics Gems", 1993, Chapter 8.1,
           pp. 404-407.
    """

    sols = []
    A = b / a
    B = c / a
    C = d / a

    # Substitute x = y - A/3 to eliminate quadratic term:
    # x^3 +px + q = 0

    sq_A = A * A
    p = 1.0/3 * (-1.0/3*sq_A + B)
    q = 1.0/2 * (2.0/27*A*sq_A - 1.0/3*A*B + C)

    # Use Cardano's formula
    cb_p = p * p * p
    D = q * q + cb_p

    if np.isclose(D, 1e-12):
        if np.isclose(q, 1e-12):  # One triple solution
            sols.append(0)
        else:  # One single and one double solution
            u = np.cbrt(-q)
            sols.append(2 * u)
            sols.append(-u)

    elif D < 0:  # Casus irreducibilis: three real solutions
        phi = 1.0/3 * np.arccos(-q/np.sqrt(-cb_p))
        t = 2*np.sqrt(-p)
        s0 = t*np.cos(phi)
        s1 = -t*np.cos(phi + np.pi/3)
        s2 = -t*np.cos(phi - np.pi/3)
        sols = [s0, s1, s2]
    else:  # one real solution
        sqrt_D = np.sqrt(D)
        u = np.cbrt(sqrt_D - q)
        v = - np.cbrt(sqrt_D + q)
        sols = [u + v]
    # end of if-else

    sub = 1.0/3*A
    for i, v in enumerate(sols):
        sols[i] -= sub

    return sols


def rotate(ap, rp, angle, tlen=None):
    """Rotate a point clockwise with respect to the given axis.

    Args:
        ap : QPointF
            Axis point.
        rp : QPointF
            Point to be rotated.
        angle: float
            Angle in degree
        tlen : float (optional)
            Target length of the rotated point
            from the axis point.

    Returns:
        tp : QPointF
             Transformed point (i.e., rotated and scaled).

    Examples:
        >>> v1 = QPointF(1, 0)
        >>> org = QPointF(0, 0)
        >>> rotate(org, v1, 90)
        PyQt5.QtCore.QPointF(0.0, 1.0)

        >>> rotate(org, v1, 90, tlen=2.0)
        PyQt5.QtCore.QPointF(0.0, 2.0)

        >>> v1 = QPointF(1, 1)
        >>> rotate(org, v1, 45)
        PyQt5.QtCore.QPointF(0.0, 1.4142135623730951)
    """
    transform = QTransform()
    v = rp - ap
    len_v = length(v)

    if len_v == 0:
        return rp

    if not tlen:
        tlen = len_v

    transform.translate(ap.x(), ap.y())
    transform.rotate(angle)
    transform.scale(tlen/len_v, tlen/len_v)
    transform.translate(-ap.x(), -ap.y())
    tp = transform.map(rp)  # Point which has been transformed
    return tp


def reflect_yaxis(ap, rp):
    """Reflect a point with respect to the y-axis.

    Args:
        ap : QPointF
            Axis point.
        rp : QPointF
            Point to be reflected.

    Returns:
        tp : QPointF
             Transformed point.
    """

    # Temporary x-coord. (tx)
    tx = rp.x() - ap.x()  # Move the x-coord of origin to that of the axis point (ap).

    # New x-coord. (nx)
    nx = -tx + ap.x()  # Reflect and find the coordinates to the original origin.
    return QPointF(nx, rp.y())


def reflect_xaxis(ap, rp):
    """Reflect a point with respect to the x-axis.

    Args:
        ap : QPointF
            Axis point.
        rp : QPointF
            Point to be reflected.

    Returns:
        tp : QPointF
             Transformed point.
    """

    # Temporary y-coord. (ty)
    ty = rp.y() - ap.y()  # Move the x-coord of origin to that of the axis point (ap).

    # New x-coord. (ny)
    ny = -ty + ap.y()  # Reflect and find the coordinates to the original origin.
    return QPointF(rp.x(), ny)


from qtpy.QtCore import Qt
from qtpy.QtCore import QObject
from qtpy.QtWidgets import QUndoView
from qtpy.QtWidgets import QUndoStack
from qtpy.QtGui import QKeySequence

from nezzle.history.commands import MoveByMouseCommand
from nezzle.history.commands import MoveByKeyCommand
from nezzle.history.commands import ConvertNodeCommand
from nezzle.history.commands import ConvertLinkCommand


class History(QObject):

    def __init__(self, parent):
        super().__init__(parent=parent)
        self._undo_stack = QUndoStack(self)

    def on_move_items_by_mouse(self, items):
        self._undo_stack.push(MoveByMouseCommand(items))

    def on_move_items_by_key(self, items):
        self._undo_stack.push(MoveByKeyCommand(items))

    def on_convert_links(self, net, old_items, new_items):
        self._undo_stack.push(ConvertLinkCommand(net, old_items, new_items))

    def on_convert_nodes(self, net, old_items, new_items):
        self._undo_stack.push(ConvertNodeCommand(net, old_items, new_items))

    def undo(self):
        self._undo_stack.undo()

    def redo(self):
        self._undo_stack.redo()

    @property
    def stack(self):
        return self._undo_stack



from nezzle.utils.etc import *
from nezzle.utils.math import *
from nezzle.utils.meta import *
from nezzle.utils.networkx import *
from nezzle.utils.triggerdict import *
from nezzle.utils.search import *
from nezzle.utils.module import *



# -*- coding: utf-8 -*-

from qtpy import QtWidgets
from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from qtpy.QtGui import QColor

from nezzle import fileio
from nezzle.graphics import LabelClassFactory
from nezzle.graphics import LinkClassFactory
from nezzle.graphics import ArrowClassFactory
from nezzle.graphics import NodeClassFactory
from nezzle.graphics import Network

app = QtWidgets.QApplication([])


def test_copy_node():
    NodeClass = NodeClassFactory.create("CIRCLE_NODE")
    n1 = NodeClass("ID1", 5, pos=QPointF(10, -20))
    n1["FILL_COLOR"] = '#FFEEAA'
    n2 = n1.copy()

    assert n1.iden == n2.iden
    assert n1["FILL_COLOR"] == n2["FILL_COLOR"]
    assert n1.pos() == n2.pos()
    assert n1.to_dict() == n2.to_dict()

    NodeClass = NodeClassFactory.create("ELLIPSE_NODE")
    n1 = NodeClass("ID1", 100, 20, pos=QPointF(10, -20))
    n1["FILL_COLOR"] = Qt.black
    n2 = n1.copy()

    assert n1.iden == n2.iden
    assert n1["FILL_COLOR"] == n2["FILL_COLOR"]
    assert n1.pos() == n2.pos()
    assert n1.to_dict() == n2.to_dict()


def test_copy_link():

    net = Network("ID_NETWORK")

    NodeClass = NodeClassFactory.create("CIRCLE_NODE")
    n1 = NodeClass("ID1", 5, pos=QPointF(10, -20))
    n2 = NodeClass("ID2", 5, pos=QPointF(10, 20))
    net.add_node(n1)
    net.add_node(n2)

    ArrowClass = ArrowClassFactory.create("TRIANGLE")
    head = ArrowClass()

    LinkClass = LinkClassFactory.create('CURVED_LINK')
    link1 = LinkClass(iden="ID3", source=n1, target=n2, width=3, head=head)
    link1["FILL_COLOR"] = QColor(10, 100, 10)
    link1['WIDTH'] = 5

    link2 = link1.copy()

    net.add_link(link1)
    net.add_link(link2)

    assert link1.iden == link2.iden
    assert link1.source == link2.source
    assert link1.target == link2.target
    assert link1["FILL_COLOR"] == link2["FILL_COLOR"]
    assert link1['WIDTH'] == link2['WIDTH']
    assert link1.to_dict() == link2.to_dict()


def test_copy_label():
    NodeClass = NodeClassFactory.create("CIRCLE_NODE")
    n1 = NodeClass("ID1", 5, pos=QPointF(10, -20))

    LabelClass = LabelClassFactory.create("TEXT_LABEL")
    label1 = LabelClass(parent=n1, text="This is text labels")
    label2 = label1.copy()

    assert label1.text == label2.text
    assert label1.font.toString() == label2.font.toString()
    assert label1.pos() == label2.pos()
    assert label1.to_dict() == label2.to_dict()


def test_copy_network():
    net1 = fileio.read_network("jkwon_egfr_pathway.sif")
    net2 = net1.copy()

    assert set(net1.nodes.keys()) == set(net2.nodes.keys())
    assert set(net1.links.keys()) == set(net2.links.keys())
    assert net1.to_dict() == net2.to_dict()



import random
import math

import networkx as nx

from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from qtpy.QtGui import QColor
from qtpy.QtGui import QFont

import nezzle
from nezzle.graphics import NodeClassFactory
from nezzle.graphics import LinkClassFactory
from nezzle.graphics import LabelClassFactory
from nezzle.graphics import ArrowClassFactory
from nezzle.graphics import Network
from nezzle.graphics import SelfloopLink
from nezzle.constants import DEFAULT_SCENE_WIDTH, DEFAULT_SCENE_HEIGHT
from nezzle.utils.math import rotate, dist, internal_division


def to_graphics(dg, iden, no_link_type=False):
    if not isinstance(dg, nx.DiGraph):
        raise TypeError("NetworkX.DiGraph should be given, not %s"%(type(dg)))

    net = Network(iden)
    #net.scene.setBackgroundBrush(QColor(0, 0, 0, 0))
    net.scene.setBackgroundBrush(Qt.transparent)

    NodeClass = NodeClassFactory.create("ELLIPSE_NODE")
    nodes = {}
    counter_node = 0
    counter_link = 0

    # Set constants with default values
    scene_width = DEFAULT_SCENE_WIDTH
    scene_height = DEFAULT_SCENE_HEIGHT

    # Node color
    color_node = Qt.white
    half_width = scene_width / 2
    half_height = scene_height / 2

    range_x = (-scene_width / 4, scene_width / 4)
    range_y = (-scene_height / 4, scene_height / 4)

    # Node size
    width = 50
    height = 35

    # Link types
    str_act = '+'
    str_inh = '-'

    for str_src, str_tgt, link_data in dg.edges(data=True):
        str_link_type = None
        head = None

        if not no_link_type:
            if 'HEAD' in link_data:
                if isinstance(link_data['HEAD'], dict):
                    attr_head = link_data.pop('HEAD')
                    ArrowClass = ArrowClassFactory.create(attr_head['TYPE'])
                    head = ArrowClass.from_dict(attr_head)
                    # if dict_head['TYPE'] == "TRIANGLE":
                    #     str_link_type = '+'
                    # elif dict_head['TYPE'] == "HAMMER":
                    #     str_link_type = '-'
                else:
                    str_link_type = link_data.pop('HEAD')
            elif 'SIGN' in link_data:
                sign_link = link_data['SIGN']
                if sign_link > 0:
                    str_link_type = str_act
                elif sign_link < 0:
                    str_link_type = str_inh
                else:
                    raise ValueError("Undefined link sign: %s"%(sign_link))

        if not head and not no_link_type \
           and (str_link_type not in (str_act, str_inh)):
            raise ValueError("Undefined link type: %s"%(str_link_type))

        if 'POS_X' in dg.nodes[str_src]:
            sx = dg.nodes[str_src]['POS_X']
        else:
            sx = half_width + random.uniform(*range_x)

        if 'POS_Y' in dg.nodes[str_src]:
            sy = dg.nodes[str_src]['POS_Y']
        else:
            sy = half_height + random.uniform(*range_y)

        if 'POS_X' in dg.nodes[str_tgt]:
            tx = dg.nodes[str_tgt]['POS_X']
        else:
            tx = half_width + random.uniform(*range_x)

        if 'POS_Y' in dg.nodes[str_tgt]:
            ty = dg.nodes[str_tgt]['POS_Y']
        else:
            ty = half_height + random.uniform(*range_y)

        if str_src in nodes:
            src = nodes[str_src]
        else:
            counter_node += 1
            src = NodeClass(str_src, width=width, height=height,
                            pos=QPointF(sx, sy))

            if "FILL_COLOR" not in dg.nodes[str_src]:
                src["FILL_COLOR"] = color_node

            if 'BORDER_COLOR' not in dg.nodes[str_src]:
                src['BORDER_COLOR'] = Qt.darkGray

            src.update(dg.nodes[str_src])
            nodes[str_src] = src
        # end of else

        if str_tgt in nodes:
            tgt = nodes[str_tgt]
        else:
            counter_node += 1
            tgt = NodeClass(str_tgt, width=width, height=height,
                            pos=QPointF(tx, ty))

            if "FILL_COLOR" not in dg.nodes[str_tgt]:
                tgt["FILL_COLOR"] = color_node

            if 'BORDER_COLOR' not in dg.nodes[str_tgt]:
                tgt['BORDER_COLOR'] = Qt.darkGray

            tgt.update(dg.nodes[str_tgt])
            nodes[str_tgt] = tgt
        # end of else

        counter_link += 1

        # Add head
        if not head:  # Head can created according to head information.
            if no_link_type:
                ArrowClass = None
            elif str_link_type == '+':
                head_type = "TRIANGLE"
                ArrowClass = ArrowClassFactory.create(head_type)
            elif str_link_type == '-':
                head_type = "HAMMER"
                ArrowClass = ArrowClassFactory.create(head_type)
            else:
                pass  # This logic is processed just below.

            if ArrowClass:
                head = ArrowClass()

        # Add link with head
        if str_src == str_tgt:  # Self-loop link
            LinkClass = LinkClassFactory.create('SELFLOOP_LINK')
            iden = "%s%s%s" % (str_src, str_link_type, str_src)
            link = LinkClass(iden=iden,
                             name=str_link_type,
                             node=src,
                             head=head)

            if "FILL_COLOR" not in link_data:
                link["FILL_COLOR"] = QColor(100, 100, 100, 100)

            # Update extra data in nezzle.graphics.Link object.
            link.update(link_data)
        else:
            LinkClass = LinkClassFactory.create('CURVED_LINK')
            iden = "%s%s%s" % (str_src, str_link_type, str_tgt)
            link = LinkClass(iden=iden,
                             name= str_link_type,
                             source=src, target=tgt,
                             head=head)

            if "FILL_COLOR" not in link_data:
                link["FILL_COLOR"] = Qt.black

            # Update extra data in nezzle.graphics.Link object.
            link.update(link_data)
        # end of else

        src.add_link(link)
        tgt.add_link(link)
        net.add_link(link)
    # end of for : reading each line of SIF file

    # Add nodes and labels in network
    font = QFont()
    font.setFamily("Tahoma")
    font.setPointSize(10)
    LabelClass = LabelClassFactory.create("TEXT_LABEL")
    for str_name, node in nodes.items():
        net.add_node(node)
        label = LabelClass(node, str_name)
        label.font = font
        rect = label.boundingRect()
        label.setPos(-rect.width()/2, -rect.height()/2)
        net.add_label(label)
        nodes[str_name] = node

    # Make the two links of interconnected nodes curved.
    for src, tgt, attr in net.nxdg.edges(data=True):
        if net.nxdg.has_edge(tgt, src):
            if src == tgt:  # Skip selfloops
                continue

            link = attr['GRAPHICS']
            mid = internal_division(link.pos_src, link.pos_tgt, 0.5, 0.5)
            d = dist(link.pos_src, mid)/math.cos(math.pi/4)
            cp = rotate(link.pos_src, mid, -30, d)
            link.ctrl_point.setPos(cp)

    return net


def to_networkx(net):
    if not isinstance(net, nezzle.graphics.Network):
        raise TypeError("nezzle.graphics.Network should be given, not %s"%(type(net)))

    dg = nx.DiGraph()
    dg.name = net.iden

    for iden, link in net.links.items():

        if isinstance(link, SelfloopLink):
            src = tgt = link.node
        else:
            src = link.source
            tgt = link.target

        if src.iden not in dg.nodes:
            dg.add_node(src.iden)
            dg.nodes[src.iden].update(src.to_dict())

        if tgt.iden not in dg.nodes:
            dg.add_node(tgt.iden)
            dg.nodes[tgt.iden].update(tgt.to_dict())


        dg.add_edge(src.iden, tgt.iden)
        link_data = dg.edges[src.iden, tgt.iden]
        link_data.update(link.to_dict())

        # Set sign information if head exists.
        if link.head:
            sign_link = 0
            if link.head.TYPE == "TRIANGLE":
                sign_link = +1
            elif link.head.TYPE == "HAMMER":
                sign_link = -1

            link_data['SIGN'] = sign_link

    return dg



import numpy as np
import networkx as nx


def layout(net, layout_func, scale=None, center=None):
    if not scale:
        coords = np.zeros((len(net.nodes), 2), dtype=np.float)
        for i, (iden, attr) in enumerate(net.nxdg.nodes_iter(data=True)):
            node = attr['GRAPHICS']
            coords[i, :] = node.x(), node.y()

        size_coords = coords.max(axis=0) - coords.min(axis=0)
        scale = np.ceil(np.max(size_coords))

    res = layout_func(net.nxdg, scale=scale, center=center)

    for iden, pos in res.items():
        node = net.nxdg.node[iden]['GRAPHICS']
        node.setX(pos[0])
        node.setY(pos[1])





# -*- coding: utf-8 -*-

def Singleton(_class):
    class __Singleton(_class):
        __instance = None

        def __new__(cls, *args, **kwargs):
            if cls.__instance is None:
                cls.__instance = super().__new__(cls, *args, **kwargs)

                # Creation and initialization of '__initialized'
                cls.__instance.__initialized = False
            # end of if
            return cls.__instance

        def __init__(self, *args, **kwargs):
            if self.__initialized:
                return

            super().__init__(*args, **kwargs)
            self.__initialized = True

        def __repr__(self):
            return '<{0} Singleton object at {1}>'.format(
                _class.__name__, hex(id(self)))

        def __str__(self):
            return super().__str__()
    # end of def class

    __Singleton.__name__ = _class.__name__
    return __Singleton




from collections import UserDict


class TriggerDict(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._triggers_set = dict()
        self._triggers_get = dict()

    def __setitem__(self, key, value):
        self.set(key, value)

    def __getitem__(self, key):
        return self.get(key)

    def set_trigger(self, key, func, when='set'):
        if when == 'set':
            self._triggers_set[key] = func
        elif when == 'get':
            self._triggers_get[key] = func
        else:
            ValueError("You should choose 'set' or 'get' "
                       "for the timing of trigger, not '%s.'"%(when))

    def set(self, key, value, trigger=True):
        if key not in self._triggers_set or not trigger:
            self.data[key] = value
            return

        if trigger:
            func = self._triggers_set[key]
            self.data[key] = func(key, value)

    def get(self, key, trigger=True):
        if key not in self._triggers_get or not trigger:
            return self.data.get(key)

        if trigger:
            func = self._triggers_get[key]
            value = self.data.get(key)
            return func(key, value)






def _set_displacement(pos, dp):
    dp_x = dp
    dp_y = dp

    if 'top' in pos:
        dp_y *= -1

    if 'left' in pos:
        dp_x *= -1

    return dp_x, dp_y


def _align_label_center(net):
    for label in net.labels:
        rect = label.boundingRect()
        label.setPos(-rect.width() / 2, -rect.height() / 2)


def align_label(net, pos='bottomright', dp=7):
    """
    Align labels with the position relative to their parent nodes
    The alignment should be updated after changing fonts,
    because setting new fonts changes the size of labels, which invalidates
    the previous alignment.
    """

    if pos == 'center':
        return _align_label_center(net)

    dp_x, dp_y = _set_displacement(pos, dp)

    for label in net.labels:
        node = label.parent
        #labels.setPos(nodes._radius + dp_x, nodes._radius + dp_y)
        label.setPos(dp_x, dp_y)







import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtCore import QLineF
from qtpy.QtCore import QRectF
from qtpy.QtGui import QBrush, QColor, QPainterPath
from qtpy.QtWidgets import QGraphicsItem

from nezzle.graphics.links.straightlink import StraightLink
from nezzle.graphics.links.controlpoint import ControlPoint

from nezzle.utils import dot
from nezzle.utils import internal_division
from nezzle.utils import length
from nezzle.graphics import quadbezier
from nezzle.graphics.mixins import Lockable


@Lockable
class CurvedLink(StraightLink):

    ITEM_TYPE = 'CURVED_LINK'

    def __init__(self, *args, **kwargs):
        self._t_head = 0
        super().__init__(*args, **kwargs)

        self._attr.set_trigger('CP_POS_X',
                               self._trigger_set_cp_pos_x,
                               when='set')

        self._attr.set_trigger('CP_POS_Y',
                               self._trigger_set_cp_pos_y,
                               when='set')


        # To avoid repetitive memory allocation, make this variable as a member.
        self._arr_t = np.arange(1, 0.5, -0.001, dtype=np.float64)

    @property
    def pos_ctrl(self):
        return self._ctrl_point.pos()

    @property
    def ctrl_point(self):
        return self._ctrl_point

    def _trigger_set_cp_pos_x(self, key, value):
        self._ctrl_point.setX(value)
        return value

    def _trigger_set_cp_pos_y(self, key, value):
        self._ctrl_point.setY(value)
        return value

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSelectedChange:
            self._ctrl_point.update()

        return super().itemChange(change, value)

    def update(self):
        self.update_ctrl_points()
        super().update()

    def update_ctrl_points(self):
        cp = self._ctrl_point
        if self.is_node_selected():
            cp.setVisible(False)
        else:
            cp.setVisible(True)

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)

        if self.isSelected():
            painter.setBrush(Qt.red)
            painter.drawEllipse(-2.5, -2.5, 5, 5)
            painter.setPen(QColor(50, 50, 50, 100))

            if not self.is_node_selected():
                # Draw control lines
                painter.drawLine(self.pos_ctrl, self.pos_src)
                painter.drawLine(self.pos_ctrl, self.pos_tgt)

    def is_straight(self):
        v1 = self.pos_src - self.ctrl_point.pos()
        v2 = self.pos_tgt - self.ctrl_point.pos()
        len_v1 = length(v1)
        len_v2 = length(v2)
        dot_v1v2 = dot(v1, v2)

        if len_v1 <= 1e-8 or len_v2 <= 1e-8:
            return True

        try:
            inner_product = np.clip(dot_v1v2/(len_v1*len_v2), -1, 1)
            control_angle = np.arccos(inner_product)
        except ZeroDivisionError:
            return True
            
        return np.isclose(control_angle, 0, atol=0.1) \
               or np.isclose(control_angle, np.pi, atol=1e-1)

    def initialize(self):
        self._identify_pos()
        self._create_control_items()
        self._create_subpoints()
        self._create_path()
        self._update_bounding_rect()

    def _update_bounding_rect(self):
        super()._update_bounding_rect()

        rect_ctrl_src = QRectF(self.pos_ctrl, self.pos_src)
        rect_ctrl_tgt = QRectF(self.pos_ctrl, self.pos_tgt)

        rect = self._bounding_rect
        rect = rect.united(rect_ctrl_src)
        rect = rect.united(rect_ctrl_tgt)
        self._bounding_rect = rect

    def _create_control_items(self):
        mid = internal_division(self.pos_src, self.pos_tgt, 0.5, 0.5)
        self._ctrl_point = ControlPoint(parent=self, pos=mid)

    def _create_subpoints(self):
        self.cps = 3 * [None]
        self.cps[0] = self.pos_src
        self.cps[2] = self.pos_tgt

        self._qps_top = []
        self._qps_bottom = []

        self._dps_top = []
        self._dps_bottom = []

    def _identify_head(self):
        CurvedLink._identify_head_pos(self)
        CurvedLink._calculate_head_angle(self)
        super()._create_head_path()

    def _identify_head_pos(self):
        offset = self._calculate_head_offset()

        p1 = self.pos_src
        p2 = self.pos_tgt
        pc = self.ctrl_point.pos()

        x = np.array([p1.x(), pc.x(), p2.x()], dtype=np.float64)
        y = np.array([p1.y(), pc.y(), p2.y()], dtype=np.float64)

        arclen = quadbezier.arc_length(x, y, self._arr_t, 1)
        rchange = np.abs(arclen - offset) / offset

        ix = np.argmax(rchange < 5e-2)
        t = self._arr_t[ix]
        self._t_head = t

        ph = (1-t)**2*p1 + 2*(1-t)*t*pc + t**2*p2
        self.pos_head.setX(ph.x())
        self.pos_head.setY(ph.y())

    def _calculate_head_offset(self):
        v = self.pos_ctrl - self.pos_tgt
        try:
            angle_rad = np.arccos(v.x() / length(v))
        except ZeroDivisionError:
            return self.head.offset + self.head.height

        radius = self.target.calculate_radius(angle_rad)
        return radius + self.head.offset + self.head.height

    def _calculate_head_angle(self):
        self._angle_head = -QLineF(self.pos_ctrl, self.pos_head).angle()
        self._head_transform.angle = self._angle_head

    def _identify_curve_points(self):
        # Control points for quadratic Bezier curve
        cps = [self.pos_src,
               self.ctrl_point.pos(),
               self.pos_tgt]

        if self.head:
            if self.is_straight():
                StraightLink._identify_head(self)
                return
            else:
                CurvedLink._identify_head(self)
                cps[2] = self.pos_head

        sps = quadbezier.identify_sps(cps)
        qps_top, qps_bottom = quadbezier.identify_qps(sps, self.width)
        dps_top, dps_bottom = quadbezier.identify_dps(sps, self.width)

        self._qps_top = qps_top
        self._qps_bottom = qps_bottom
        self._dps_top = dps_top
        self._dps_bottom = dps_bottom

    def _create_curve_path(self):

        self._identify_curve_points()
        self._path_paint = QPainterPath()
        # self._path_paint.clear()  # supported from Qt 5.13
        self._path_paint.setFillRule(Qt.WindingFill)

        self._path_paint.moveTo(self._qps_top[0])
        for i in range(len(self._dps_top)):
            self._path_paint.quadTo(self._dps_top[i], self._qps_top[i + 1])

        if self.head:
            self._path_paint.connectPath(self._path_head)
        else:
            self._path_paint.lineTo(self._qps_bottom[-1])

        for i in range(len(self._dps_bottom) - 1, -1, -1):
            self._path_paint.quadTo(self._dps_bottom[i],
                                    self._qps_bottom[i])

        self._path_paint.lineTo(self._qps_top[0])

    def _create_path(self):
        self._identify_pos()
        if self.is_straight():
            return StraightLink._create_path(self)

        try:
            self._create_curve_path()
        except FloatingPointError:
            super(StraightLink, self)._create_path()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './nezzle/ui/ui_opennetworkdialog.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_OpenNetworkDialog(object):
    def setupUi(self, OpenNetworkDialog):
        OpenNetworkDialog.setObjectName("OpenNetworkDialog")
        OpenNetworkDialog.resize(591, 470)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(OpenNetworkDialog.sizePolicy().hasHeightForWidth())
        OpenNetworkDialog.setSizePolicy(sizePolicy)
        OpenNetworkDialog.setWindowOpacity(2.0)
        OpenNetworkDialog.setSizeGripEnabled(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(OpenNetworkDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_fpath = QtWidgets.QLabel(OpenNetworkDialog)
        self.label_fpath.setObjectName("label_fpath")
        self.horizontalLayout.addWidget(self.label_fpath)
        self.ui_filePathEdit = QtWidgets.QLineEdit(OpenNetworkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_filePathEdit.sizePolicy().hasHeightForWidth())
        self.ui_filePathEdit.setSizePolicy(sizePolicy)
        self.ui_filePathEdit.setObjectName("ui_filePathEdit")
        self.horizontalLayout.addWidget(self.ui_filePathEdit)
        self.ui_openButton = QtWidgets.QPushButton(OpenNetworkDialog)
        self.ui_openButton.setObjectName("ui_openButton")
        self.horizontalLayout.addWidget(self.ui_openButton)
        self.ui_reloadButton = QtWidgets.QPushButton(OpenNetworkDialog)
        self.ui_reloadButton.setObjectName("ui_reloadButton")
        self.horizontalLayout.addWidget(self.ui_reloadButton)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_network_name = QtWidgets.QLabel(OpenNetworkDialog)
        self.label_network_name.setObjectName("label_network_name")
        self.horizontalLayout_2.addWidget(self.label_network_name)
        self.ui_networkNameEdit = QtWidgets.QLineEdit(OpenNetworkDialog)
        self.ui_networkNameEdit.setObjectName("ui_networkNameEdit")
        self.horizontalLayout_2.addWidget(self.ui_networkNameEdit)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.ui_labelLinkHeadMapping = QtWidgets.QLabel(OpenNetworkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_labelLinkHeadMapping.sizePolicy().hasHeightForWidth())
        self.ui_labelLinkHeadMapping.setSizePolicy(sizePolicy)
        self.ui_labelLinkHeadMapping.setObjectName("ui_labelLinkHeadMapping")
        self.verticalLayout.addWidget(self.ui_labelLinkHeadMapping)
        self.ui_linkHeadMappingTable = QtWidgets.QTableWidget(OpenNetworkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_linkHeadMappingTable.sizePolicy().hasHeightForWidth())
        self.ui_linkHeadMappingTable.setSizePolicy(sizePolicy)
        self.ui_linkHeadMappingTable.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustIgnored)
        self.ui_linkHeadMappingTable.setAlternatingRowColors(True)
        self.ui_linkHeadMappingTable.setObjectName("ui_linkHeadMappingTable")
        self.ui_linkHeadMappingTable.setColumnCount(0)
        self.ui_linkHeadMappingTable.setRowCount(0)
        self.ui_linkHeadMappingTable.horizontalHeader().setCascadingSectionResizes(False)
        self.verticalLayout.addWidget(self.ui_linkHeadMappingTable)
        self.ui_labelLinkTailMapping = QtWidgets.QLabel(OpenNetworkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_labelLinkTailMapping.sizePolicy().hasHeightForWidth())
        self.ui_labelLinkTailMapping.setSizePolicy(sizePolicy)
        self.ui_labelLinkTailMapping.setObjectName("ui_labelLinkTailMapping")
        self.verticalLayout.addWidget(self.ui_labelLinkTailMapping)
        self.ui_linkTailMappingTable = QtWidgets.QTableWidget(OpenNetworkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ui_linkTailMappingTable.sizePolicy().hasHeightForWidth())
        self.ui_linkTailMappingTable.setSizePolicy(sizePolicy)
        self.ui_linkTailMappingTable.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustIgnored)
        self.ui_linkTailMappingTable.setAlternatingRowColors(True)
        self.ui_linkTailMappingTable.setObjectName("ui_linkTailMappingTable")
        self.ui_linkTailMappingTable.setColumnCount(0)
        self.ui_linkTailMappingTable.setRowCount(0)
        self.ui_linkTailMappingTable.horizontalHeader().setCascadingSectionResizes(False)
        self.verticalLayout.addWidget(self.ui_linkTailMappingTable)
        self.ui_buttonBox = QtWidgets.QDialogButtonBox(OpenNetworkDialog)
        self.ui_buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.ui_buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.ui_buttonBox.setObjectName("ui_buttonBox")
        self.verticalLayout.addWidget(self.ui_buttonBox)

        self.retranslateUi(OpenNetworkDialog)
        self.ui_buttonBox.accepted.connect(OpenNetworkDialog.accept)
        self.ui_buttonBox.rejected.connect(OpenNetworkDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(OpenNetworkDialog)

    def retranslateUi(self, OpenNetworkDialog):
        _translate = QtCore.QCoreApplication.translate
        OpenNetworkDialog.setWindowTitle(_translate("OpenNetworkDialog", "Dialog"))
        self.label_fpath.setText(_translate("OpenNetworkDialog", "File Path"))
        self.ui_openButton.setText(_translate("OpenNetworkDialog", "Open"))
        self.ui_reloadButton.setText(_translate("OpenNetworkDialog", "Reload"))
        self.label_network_name.setText(_translate("OpenNetworkDialog", "Network Name"))
        self.ui_labelLinkHeadMapping.setText(_translate("OpenNetworkDialog", "Link Head Mapping"))
        self.ui_linkHeadMappingTable.setSortingEnabled(True)
        self.ui_labelLinkTailMapping.setText(_translate("OpenNetworkDialog", "Link Tail Mapping"))
        self.ui_linkTailMappingTable.setSortingEnabled(True)




import traceback

import numpy as np

from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from qtpy.QtCore import QLineF
from qtpy.QtCore import QRectF
from qtpy.QtGui import QVector2D
from qtpy.QtGui import QPen, QBrush, QColor
from qtpy.QtGui import QPainterPath
from qtpy.QtWidgets import QGraphicsItem

from nezzle.graphics.links.straightlink import StraightLink
from nezzle.graphics.links.controlpoint import HorizontalConnectorControlPoint
from nezzle.graphics.links.controlpoint import VerticalConnectorControlPoint

from nezzle.utils import angle
from nezzle.utils import dist
from nezzle.utils import dot
from nezzle.utils import internal_division
from nezzle.utils import normal_vector
from nezzle.utils import length
from nezzle.utils import rotate
from nezzle.graphics import quadbezier
from nezzle.graphics.mixins import Lockable

np.seterr('raise')


class ElbowLink(StraightLink):
    ITEM_TYPE = 'ELBOW_LINK'

    def __init__(self, *args, **kwargs):
        self._ctrl_points = []
        self._t_head = 0

        # To avoid repetitive memory allocation, make this variable as a member.
        self._arr_t = np.arange(1, 0.5, -0.001, dtype=np.float64)
        super().__init__(*args, **kwargs)  # initialize() is called in the hierarchy of super().__init__

        for i in range(3):
            self._attr.set_trigger(f'CP{i}_POS_X',
                                   self._trigger_set_cp_pos_x,
                                   when='set')

            self._attr.set_trigger(f'CP{i}_POS_Y',
                                   self._trigger_set_cp_pos_y,
                                   when='set')


    @property
    def ctrl_points(self):
        return self._ctrl_points

    def _trigger_set_cp_pos_x(self, key, value):
        ix = int(key[2])
        cp = self._ctrl_points[ix]
        old_selected = cp.isSelected()
        cp.setSelected(True)
        cp.setX(value)
        cp.setSelected(old_selected)
        return value

    def _trigger_set_cp_pos_y(self, key, value):
        ix = int(key[2])
        cp = self._ctrl_points[ix]
        old_selected = cp.isSelected()
        cp.setSelected(True)
        cp.setY(value)
        cp.setSelected(old_selected)
        return value

    def initialize(self):
        self._identify_pos()
        self._create_connectors()
        self._create_control_items()
        self._create_subpoints()
        self._create_path()
        self._update_bounding_rect()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSelectedHasChanged:
            self.update()

        return super().itemChange(change, value)

    def update(self):
        self.update_ctrl_points()
        super().update()

    def update_ctrl_points(self):
        for cp in self._ctrl_points:
            if not cp.isSelected():
                cp.update_pos_by_connectors()
                cp.update()

            if self.is_node_selected():
                cp.setVisible(False)
            else:
                cp.setVisible(True)

        # end of for

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)

        if self.isSelected():
            painter.setBrush(Qt.red)
            painter.drawEllipse(-2.5, -2.5, 5, 5)
            painter.setPen(QColor(50, 50, 50, 100))

    def is_head_visible(self):
        if not self.head:
            return False

        v1 = self._cps[-2] - self._cps[-3]
        v2 = self.pos_head - self._cps[-3]
        ip = dot(v1, v2)  # Inner product
        return ip > 0  # Is v1.v2 = |v1||v2|cos(0) positive?

    def _create_connectors(self):
        self._pos_connectors = [QPointF(), QPointF(), QPointF(), QPointF()]

    def _create_subpoints(self):
        self._cps = []  # Central points
        self._fps = []  # Forward points
        self._bps = []  # Backward points

    def _create_control_items(self):
        raise NotImplementedError()

    def _identify_connectors_pos(self):
        raise NotImplementedError()

    def _identify_head_pos(self):
        offset = self._calculate_head_offset()

        pos_conn = self._cps[-3]
        pos_end = self._cps[-2]

        ph = internal_division(pos_conn, pos_end, dist(pos_conn, pos_end) - offset, offset)
        self.pos_head.setX(ph.x())
        self.pos_head.setY(ph.y())

    def _calculate_head_angle(self):
        # pos_vc1 = self._pos_connectors[1]
        pos_begin = self._cps[-3]
        self._angle_head = -QLineF(pos_begin, self.pos_head).angle()
        self._head_transform.angle = self._angle_head
        # print("Head Angle:", self._angle_head)

    def _identify_head(self):
        self._identify_head_pos()
        self._calculate_head_angle()
        self._create_head_path()

    def _calculate_head_offset(self):
        """
        self._cps[-1]: dummy point
        self._cps[-2]: real point (the end of node)
        self._cps[-3]: previous connector (the tail of head)
        """

        v = self._cps[-2] - self._cps[-3]

        try:
            angle_rad = np.arccos(v.x()/length(v))
        except ZeroDivisionError:
            return 0
        except Exception as err:
            print(err)
            return 0

        radius = self.target.calculate_radius(angle_rad)
        return radius + self.head.offset + self.head.height

    def _create_head_path(self):
        points = self.head.identify_points(self.pos_head,
                                             self.width,
                                             self._head_transform)
        path = QPainterPath()

        path.moveTo(points[-1])
        for pt in points[-2::-1]:
            path.lineTo(pt)

        self._path_head = path

    def _identify_elbow_points(self):
        # Update positions of source and target
        self._cps.clear()
        self._cps.append(self.pos_src)  # Dummy point

        for i, pos in enumerate(self._pos_connectors):
            self._cps.append(pos)

        self._cps.append(self.pos_tgt)

        ix_begin = 0
        ix_end = len(self._cps) - 1

        # Position of each connector with respect to the position of target
        len_cps = len(self._cps)
        for i in range(1, len_cps-1):
            pos_conn = self._cps[i]
            pos_conn_on_src = pos_conn - self.pos_src
            if self.source.contains(pos_conn_on_src):
                ix_begin = i

            break

        for i in range(len_cps-2, 1, -1):
            pos_conn = self._cps[i]
            pos_conn_on_tgt = pos_conn - self.pos_tgt
            if self.target.contains(pos_conn_on_tgt):
                ix_end = i

            break

        if ix_end - ix_begin <= 1:
            self._cps = [self._cps[0], self._cps[0], self._cps[1], self._cps[-2], self._cps[-1], self._cps[-1]]
        else:
            self._cps = [self._cps[ix_begin]] + self._cps[ix_begin:ix_end + 1] + [self._cps[ix_end]]

        # print(f"(ix_begin, ix_end)=({ix_begin}, {ix_end})")


        hw = self.width / 2  # The half of width
        len_cps = len(self._cps)

        self._fps.clear()
        for i in range(1, len_cps-1):
            p0 = self._cps[i - 1]
            p1 = self._cps[i]
            p2 = self._cps[i + 1]
            v1 = QVector2D(p1 - p0)
            v2 = QVector2D(p2 - p1)

            nv1 = hw * normal_vector(v1)  # Normal vector of v1
            nv2 = hw * normal_vector(v2)  # Normal vector of v2
            self._fps.append(p1 + (nv1 + nv2).toPointF())
        # end of for

        self._bps.clear()
        for i in range(len_cps - 2, 0, -1):
            p0 = self._cps[i + 1]
            p1 = self._cps[i]
            p2 = self._cps[i - 1]
            v1 = QVector2D(p1 - p0)
            v2 = QVector2D(p2 - p1)

            nv1 = hw * normal_vector(v1)  # Normal vector of v1
            nv2 = hw * normal_vector(v2)  # Normal vector of v2
            self._bps.append(p1 + (nv1 + nv2).toPointF())
        # end of for

    def _create_elbow_path(self):
        self._identify_elbow_points()
        self._path_paint = QPainterPath()
        self._path_paint.setFillRule(Qt.WindingFill)

        if not self._fps or not self._bps:
            return

        # Forward line
        self._path_paint.moveTo(self._fps[0])
        for i in range(1, len(self._fps) - 1):
            self._path_paint.lineTo(self._fps[i])
        # end of for

        if self.head:  # Add the head
            self._identify_head()

        if self.is_head_visible():
            self._path_paint.connectPath(self._path_head)
        else:  # if head is not visible, hide the head.
            self._path_paint.lineTo(self._fps[-1])
            self._path_paint.lineTo(self._bps[0])

        # Backward line
        for i in range(1, len(self._bps)):
            self._path_paint.lineTo(self._bps[i])
        # end of for

    def _create_path(self):
        self._identify_pos()  # Identify the position of this link
        self._identify_connectors_pos()  # Identify the positions of connectors
        try:
            # if self.is_straight():
            #     return super()._create_path()
            if self.are_nodes_close():
                return

            self._create_elbow_path()
        except Exception as err:
            traceback.print_exc()
            print(err)


@Lockable
class VerticalElbowLink(ElbowLink):

    ITEM_TYPE = 'VERTICAL_ELBOW_LINK'

    def _create_control_items(self):
        m_st = internal_division(self.pos_src, self.pos_tgt, 0.5, 0.5)
        m_sm = internal_division(self.pos_src, m_st, 0.5, 0.5)
        m_mt = internal_division(m_st, self.pos_tgt, 0.5, 0.5)

        cp0 = VerticalConnectorControlPoint(iden="CP0", parent=self, pos=m_sm)
        cp0.append_connector(self._pos_connectors[0])
        cp0.append_connector(self._pos_connectors[1])
        self._ctrl_points.append(cp0)

        cp1 = HorizontalConnectorControlPoint(iden="CP1", parent=self, pos=m_st)
        cp1.append_connector(self._pos_connectors[1])
        cp1.append_connector(self._pos_connectors[2])
        self._ctrl_points.append(cp1)

        cp2 = VerticalConnectorControlPoint(iden="CP2", parent=self, pos=m_mt)
        cp2.append_connector(self._pos_connectors[2])
        cp2.append_connector(self._pos_connectors[3])
        self._ctrl_points.append(cp2)

    def _identify_connectors_pos(self):
        pos_cp0 = self._ctrl_points[0].pos()
        pos_cp1 = self._ctrl_points[1].pos()
        pos_cp2 = self._ctrl_points[2].pos()

        con0 = self._pos_connectors[0]
        con1 = self._pos_connectors[1]
        con2 = self._pos_connectors[2]
        con3 = self._pos_connectors[3]

        con0.setX(self.pos_src.x())
        con0.setY(pos_cp0.y())

        con1.setX(pos_cp1.x())
        con1.setY(pos_cp0.y())

        con2.setX(pos_cp1.x())
        con2.setY(pos_cp2.y())

        con3.setX(self.pos_tgt.x())
        con3.setY(pos_cp2.y())


@Lockable
class HorizontalElbowLink(ElbowLink):

    ITEM_TYPE = 'HORIZONTAL_ELBOW_LINK'

    def _create_control_items(self):
        m_st = internal_division(self.pos_src, self.pos_tgt, 0.5, 0.5)
        m_sm = internal_division(self.pos_src, m_st, 0.5, 0.5)
        m_mt = internal_division(m_st, self.pos_tgt, 0.5, 0.5)

        cp0 = HorizontalConnectorControlPoint(iden="CP0", parent=self, pos=m_sm)
        cp0.append_connector(self._pos_connectors[0])
        cp0.append_connector(self._pos_connectors[1])
        self._ctrl_points.append(cp0)

        cp1 = VerticalConnectorControlPoint(iden="CP1", parent=self, pos=m_st)
        cp1.append_connector(self._pos_connectors[1])
        cp1.append_connector(self._pos_connectors[2])
        self._ctrl_points.append(cp1)

        cp2 = HorizontalConnectorControlPoint(iden="CP2", parent=self, pos=m_mt)
        cp2.append_connector(self._pos_connectors[2])
        cp2.append_connector(self._pos_connectors[3])
        self._ctrl_points.append(cp2)

    def _identify_connectors_pos(self):
        pos_cp0 = self._ctrl_points[0].pos()
        pos_cp1 = self._ctrl_points[1].pos()
        pos_cp2 = self._ctrl_points[2].pos()

        con0 = self._pos_connectors[0]
        con1 = self._pos_connectors[1]
        con2 = self._pos_connectors[2]
        con3 = self._pos_connectors[3]

        con0.setX(pos_cp0.x())
        con0.setY(self.pos_src.y())

        con1.setX(pos_cp0.x())
        con1.setY(pos_cp1.y())

        con2.setX(pos_cp2.x())
        con2.setY(pos_cp1.y())

        con3.setX(pos_cp2.x())
        con3.setY(self.pos_tgt.y())



import os
import importlib
import traceback

from qtpy.QtCore import Qt
from qtpy.QtCore import QObject
from qtpy.QtCore import Slot
from qtpy.QtCore import Signal
from qtpy.QtCore import QThread

from qtpy.QtWidgets import QWidget
from qtpy.QtWidgets import QMessageBox
from qtpy.QtWidgets import QDialog, QFileDialog
from qtpy.QtWidgets import QProgressBar

#import nezzle.utils
from nezzle.utils import reload_modules
from nezzle.utils import extract_name_and_ext
from nezzle.dialogs.progressdialog import ProgressDialog


class TaskThread(QThread):
    task_finished = Signal()
    def __init__(self, exc, nav, net):
        super().__init__()
        self._exc = exc
        self._nav = nav
        self._net = net

    def run(self):
        self._exc(self._nav, self._net)
        self.task_finished.emit()


class Worker(QObject):
    finished = Signal(object)

    def set_job(self, func, net):
        self._func_execute = func
        self._net = net

    @Slot()
    def execute(self):
        nets = self._func_execute(self._net)
        self.finished.emit(nets)


class CodeManager(QObject):
    def __init__(self, mainWindow):

        super().__init__(mainWindow)

        self.mw = mainWindow
        self._open_button = self.mw.ui_openCodeButton
        self._run_button = self.mw.ui_runButton
        self._file_path_edit = self.mw.ui_codeFilePathEdit

        self._open_button.released.connect(self.on_open_button_released)
        self._run_button.released.connect(self.on_run_button_released)
        self._run_button.setShortcut("Ctrl+R")

        self._worker = Worker()
        self._thread = QThread()
        self._worker.moveToThread(self._thread)
        self._worker.finished.connect(self.on_thread_finished)
        self._thread.started.connect(self._worker.execute)


    @property
    def file_path_edit(self):
        return self._file_path_edit

    def show_exception_message(self, err):
        print(err)
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setWindowTitle("An error has occurred")
        msg.setText(traceback.format_exc())
        msg.exec()


    @Slot()
    def on_open_button_released(self):
        try:
            dialog = QFileDialog(self.mw)
            dialog.setWindowTitle("Open a Python code")
            dialog.setAcceptMode(QFileDialog.AcceptOpen)
            dialog.setNameFilters([self.tr("Python module files (*.py)")])
            dialog.setFileMode(QFileDialog.ExistingFile)
            if dialog.exec() == QDialog.Accepted:
                fpath = dialog.selectedFiles()[0]
                fpath = fpath.strip()
                if not fpath.endswith('.py'):
                    raise RuntimeError("An inappropriate file "
                                       "for Python code: %s"%(fpath))
                # end of if
                self._file_path_edit.setText(fpath)
            # end of if
        except Exception as err:
            self.show_exception_message(err)

    @Slot()
    def on_run_button_released(self):
        try:
            self.process_code(self.file_path_edit.text())
        except Exception as err:
            self.show_exception_message(err)


    @Slot(object)
    def on_thread_finished(self, nets):
        self._thread.quit()
        if nets:
            for net in nets:
                self.mw.nt_manager.append_item(net.copy())

    def process_code(self, fpath):
        fname, fext = extract_name_and_ext(fpath)

        if not fname:
            raise RuntimeError("The path for the Python module is not valid.")

        # [REF] https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        spec = importlib.util.spec_from_file_location(fname, fpath)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        if self.mw.nt_manager.current_item:
            net = self.mw.nt_manager.current_item.data()
        else:
            net = None

        if hasattr(mod, 'create'):
            self._worker.set_job(mod.create, net.copy())
            self._thread.start()

        if hasattr(mod, 'update'):
            mod.update(self.mw.nt_manager, net)






import numpy as np

from qtpy.QtCore import QPointF
from qtpy.QtCore import QLineF
from qtpy.QtGui import QPainterPath

from nezzle.utils import dist
from nezzle.utils import length
from nezzle.utils import internal_division
from nezzle.utils import rotate
from nezzle.graphics.mixins import Lockable
from nezzle.graphics.links.baselink import TwoNodeLink
from nezzle.graphics.arrows.transform import Rotate


@Lockable
class StraightLink(TwoNodeLink):
    """Straight link class.
    """
    ITEM_TYPE = 'STRAIGHT_LINK'

    def __init__(self, *args, **kwargs):
        self._angle_head = None
        self._head_transform = Rotate()
        super().__init__(*args, **kwargs)

    def initialize(self):
        self._identify_pos()
        self._create_path()
        self._update_bounding_rect()

    def _identify_head(self):
        StraightLink._identify_head_pos(self)
        StraightLink._calculate_head_angle(self)
        super()._create_head_path()

    def _identify_head_pos(self):
        offset = self._calculate_head_offset()

        p1 = self.pos_src
        p2 = self.pos_tgt

        ph = internal_division(p1, p2, dist(p1, p2) - offset, offset)
        self.pos_head.setX(ph.x())
        self.pos_head.setY(ph.y())

    def _calculate_head_offset(self):
        v = self.pos_tgt - self.pos_src
        try:
            angle_rad = np.arccos(v.x()/length(v))
        except ZeroDivisionError:
            return 0

        radius = self.target.calculate_radius(angle_rad)
        return radius + self.head.offset + self.head.height

    def _calculate_head_angle(self):
        self._angle_head = -QLineF(self.pos_src, self.pos_tgt).angle()
        self._head_transform.angle = self._angle_head

    def _identify_pos(self):
        m = internal_division(self.source.pos(), self.target.pos(), 0.5, 0.5)
        self.setPos(m)

    def _create_path(self):
        """
        Both identifying the points and creating the path for a straight line
        are handled by this method.
        """

        self._identify_pos()
        self._path_paint = QPainterPath()

        line = QLineF(self.pos_src, self.pos_tgt)

        butt1 = self.pos_src + QPointF(0, -self.width / 2)
        butt2 = self.pos_src + QPointF(0, +self.width / 2)

        angle = -line.angle()  # QLine's angle is counter-clock wise.

        butt2_rotated = rotate(self.pos_src, butt2, angle)
        butt1_rotated = rotate(self.pos_src, butt1, angle)
        self._path_paint.moveTo(butt2_rotated)
        self._path_paint.lineTo(butt1_rotated)

        if self.head and not self.are_nodes_close():
            StraightLink._identify_head(self)
            self._path_paint.connectPath(self._path_head)
            self._path_paint.lineTo(butt2_rotated)
        else:
            head1 = self.pos_src + QPointF(line.length(), -self.width / 2)
            head2 = self.pos_src + QPointF(line.length(), +self.width / 2)

            self._path_paint.lineTo(rotate(self.pos_src, head1, angle))
            self._path_paint.lineTo(rotate(self.pos_src, head2, angle))

    def paint(self, painter, option, widget):
        super().paint(painter, option, widget)
        painter.drawPath(self._path_paint)





from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMainWindow

from nezzle.ui.ui_mainwindow import Ui_MainWindow
from nezzle.managers.singleviewmanager import SingleViewManager
from nezzle.managers.networktreemanager import NetworkTreeManager
from nezzle.managers.consoletabmanager import ConsoleTabManager
from nezzle.managers.networkmodelmanager import NetworkModelManager
from nezzle.managers.codemanager import CodeManager
from nezzle.managers.historyviewmanager import HistoryViewManager
from nezzle.menu import MenuActionHandler


class MainWindow(QMainWindow, Ui_MainWindow):
    def __init__(self):
        super().__init__()

        self.setupUi(self)   # GUI parts
        self.ui_visStyleDock.close()

        self.sv_manager = SingleViewManager(self)
        self.nm_manager = NetworkModelManager(self)
        self.nt_manager = NetworkTreeManager(self)
        self.ct_manager = ConsoleTabManager(self)
        self.code_manager = CodeManager(self)
        self.hv_manager = HistoryViewManager(self)

        # The MenuActionHandler object should be created after creating all managers.
        self.ma_handler = MenuActionHandler(self)

        self.setAttribute(Qt.WA_DeleteOnClose)

    # end of __init__

    def closeEvent(self, event):
        self.ct_manager.tab_widget.removeTab(0)
        self.ct_manager.console_widget.stop()
        event.accept()



import os
import os.path as osp
from setuptools import setup, find_packages

with open('README.md') as fin:
    lines = []
    for line in fin:
        if line.startswith("<"):
            continue 
        
        if "## Examples" in line:
            break
        
        lines.append(line)
        
    long_description = "".join(lines)


with open("nezzle/VERSION", "rt") as fin:
    version = fin.read().strip()

scripts = []
if os.name == 'nt':
    fpath_script = osp.join('scripts', 'nezzle.bat')
else:
    fpath_script = osp.join('scripts', 'nezzle')

scripts.append(fpath_script)


setup (
    name="nezzle",
    version=version,
    description="Nezzle: a programmable and interactive "
                "network visualization software",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dwgoon/nezzle",
    author="Daewon Lee",
    author_email="daewon4you@gmail.com",
    license="MIT",
    python_requires=">=3.6",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(),
    package_data={
        "": ["VERSION", "*.ui", "*.sif", "*.json"],
        "nezzle.resources": ["icon.png", "logo.png"],
    },
    entry_points={
        "console_scripts": ["nezzle = nezzle.app:main"],
    },
    scripts = scripts,
)



"""
[REF]
- http://stackoverflow.com/a/11604577
- http://stackoverflow.com/a/12144823
"""

import sys

if sys.version_info.minor > 9:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping

from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from qtpy.QtWidgets import QGraphicsItem
from qtpy.QtWidgets import QStyle
from qtpy.QtGui import QColor
from qtpy.QtGui import QPen

from nezzle.graphics.attributemapper import AttributeMapper
from nezzle.utils import TriggerDict


class MappableGraphicsItemMeta(type(QGraphicsItem),
                               type(MutableMapping)):

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)


class MappableItem(MutableMapping):
    def __init__(self, iden, *args, name=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._attr = TriggerDict()
        self._attr.set_trigger('ID', self._trigger_set_iden)
        self._attr.set_trigger('NAME', self._trigger_set_name)
        self._attr['ID'] = iden
        self._attr['NAME'] = name if name else iden

    @property
    def iden(self):
        return self._attr['ID']

    @iden.setter
    def iden(self, val):
        self._attr['ID'] = val

    @property
    def name(self):
        return self._attr['NAME']

    @name.setter
    def name(self, val):
        self._attr['NAME'] = val

    def _trigger_set_iden(self, key, value):
        self._iden = value
        return value

    def _trigger_set_name(self, key, value):
        self._name = value
        return value

    def __getitem__(self, key):
        return self._attr[key]

    def __setitem__(self, key, value):
        self._attr[key] = value

    def __delitem__(self, key):
        del self._attr[key]

    def __iter__(self):
        return iter(self._attr)

    def __len__(self):
        return len(self._attr)

    def __str__(self):
        return str(self._attr)

    def keys(self):
        return self._attr.keys()

    def values(self):
        return self._attr.values()

    def to_dict(self):
        return dict(self._attr)

    @classmethod
    def from_dict(cls, attr):
        obj = cls()
        for key, value in attr.items():
            obj[key] = value

        return obj


class MappableGraphicsItem(MappableItem,
                           QGraphicsItem,
                           metaclass=MappableGraphicsItemMeta):
    """
    A simple mappable class, which handles multiple attributes with
    its own attribute dictionary.
    """
    ITEM_TYPE = 'MAPPABLE_GRAPHICS_ITEM'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._attr['ITEM_TYPE'] = self.ITEM_TYPE


class Movable(object):
    def is_movable(self):
        return False


class GeometryChangeItem(MappableGraphicsItem,
                         Movable):

    ITEM_TYPE = 'GEOMETRY_CHANGE_ITEM'

    def __init__(self, *args, **kwargs):
        pos = None
        if "pos" in kwargs:
            pos = kwargs.pop("pos")

        super().__init__(*args, **kwargs)

        self._attr.set_trigger('POS_X', self._trigger_set_x)
        self._attr.set_trigger('POS_Y', self._trigger_set_y)

        if pos:
            self._attr['POS_X'] = pos.x()
            self._attr['POS_Y'] = pos.y()
            self._attr['_OLD_POS'] = QPointF(pos)
        else:
            self._attr['POS_X'] = 0
            self._attr['POS_Y'] = 0
            self._attr['_OLD_POS'] = QPointF(0, 0)

        self.setFlags(QGraphicsItem.ItemIsMovable
                      | QGraphicsItem.ItemSendsGeometryChanges)

    def _trigger_set_x(self, key, value):
        self.setX(value)
        return value

    def _trigger_set_y(self, key, value):
        self.setY(value)
        return value

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionHasChanged:
            pos = value
            self._attr.set('POS_X', pos.x(), trigger=False)
            self._attr.set('POS_Y', pos.y(), trigger=False)

        return super().itemChange(change, value)


    def gather_children(self):
        children = self.childItems()
        return children

    def update_children_old_positions(self):
        children = self.gather_children()
        if children:
            for child in children:
                if child.is_movable():
                    child["_OLD_POS"] = child.pos()


class PainterOptionItem(GeometryChangeItem):

    ITEM_TYPE = 'PAINTER_OPTION_ITEM'

    _attr_map = AttributeMapper()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._attr.set_trigger("BORDER_WIDTH", self._trigger_set_border_width)
        self._attr.set_trigger('BORDER_COLOR', self._trigger_set_border_color)
        self._attr.set_trigger('BORDER_JOIN', self._trigger_set_border_join)
        self._attr.set_trigger('BORDER_LINE', self._trigger_set_border_line)
        self._attr.set_trigger("FILL_COLOR", self._trigger_set_fill_color)

        self._pen = QPen()
        self._pen.setBrush(Qt.transparent)
        self._pen.setJoinStyle(Qt.MiterJoin)

        self._color = QColor()

    @property
    def attr_map(self):
        return __class__._attr_map

    def _invalidate(self):
        if self.scene():
            self.scene().invalidate()

    def _trigger_set_border_width(self, key, value):
        width = int(value)
        self._pen.setWidth(width)
        self._invalidate()
        return width

    def _trigger_set_border_color(self, key, value):
        color = QColor(value)
        self._pen.setColor(color)
        self._invalidate()
        return color.name(QColor.HexArgb)

    def _trigger_set_border_join(self, key, value):
        if isinstance(value, str):
            join_style = self.attr_map.to_qt('BORDER_JOIN', value)
            self._pen.setJoinStyle(join_style)
            return value
        elif isinstance(value, Qt.JoinStyle):
            self._pen.setJoinStyle(value)
            return self.attr_map.to_str('BORDER_JOIN', value)
        else:
            err_msg = "It is not an border join option: %s"%(value)
            raise AttributeError(err_msg)

        self._invalidate()

        return value

    def _trigger_set_border_line(self, key, value):
        if isinstance(value, str):
            line_type = self.attr_map.to_qt('BORDER_LINE', value)
            self._pen.setStyle(line_type)
            return value
        elif isinstance(value, Qt.PenStyle):
            self._pen.setStyle(value)
            return self.attr_map.to_str('BORDER_LINE', value)
        else:
            err_msg = "It is not an border line type: %s"%(value)
            raise AttributeError(err_msg)

        self._invalidate()

        return value

    def _trigger_set_fill_color(self, key, value):
        self._color = QColor(value)
        self._invalidate()
        return self._color.name(QColor.HexArgb)

    def _set_border_options(self, painter):
        painter.setPen(self._pen)

    def _set_fill_options(self, painter):
        painter.setBrush(self._color)

    def _set_painter_options(self, painter):
        painter.setPen(Qt.NoPen)
        self._set_border_options(painter)
        self._set_fill_options(painter)

    def paint(self, painter, option, widget):

        painter.setClipRect(option.exposedRect)

        if option.state & QStyle.State_Selected:
            pen = QPen(self._pen)
            pen.setBrush(Qt.green)
            painter.setPen(pen)
            painter.setBrush(Qt.green)
        else:
            self._set_painter_options(painter)






import os

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QFileDialog
from qtpy.QtWidgets import QMessageBox
from qtpy.QtGui import QDoubleValidator
from qtpy.QtGui import QIntValidator
from qtpy.QtCore import Slot

from nezzle.ui.ui_exportimagedialog import Ui_ExportImageDialog
from nezzle.utils import extract_name_and_ext
#from nezzle.constants import DEFAULT_SCENE_WIDTH, DEFAULT_SCENE_HEIGHT

class ExportImageDialog(QDialog, Ui_ExportImageDialog):

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.setWindowTitle("Export an image")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self.setFixedSize(self.width(), self.height())

        validator_int_percent = QIntValidator(0, 100)
        validator_int_nonnegative = QIntValidator()
        validator_int_nonnegative.setBottom(0)

        #validator_double_nonnegative = QDoubleValidator()
        #validator_double_nonnegative.setBottom(0.0)

        # Set the bounds of quality slider
        self.ui_qualitySlider.setMinimum(0)
        self.ui_qualitySlider.setMaximum(100)

        # Set validators
        self.ui_qualityEdit.setValidator(validator_int_percent)

        self.ui_scaleWidthEdit.setValidator(validator_int_nonnegative)
        self.ui_scaleHeightEdit.setValidator(validator_int_nonnegative)

        self.ui_dpiWidthEdit.setValidator(validator_int_nonnegative)
        self.ui_dpiHeightEdit.setValidator(validator_int_nonnegative)

        # Set default values
        self.ui_transparencyCheck.setChecked(True)

        self.ui_qualitySlider.setValue(100)
        self.ui_qualityEdit.setText('100')

        self.ui_scaleWidthEdit.setText('200')
        self.ui_scaleHeightEdit.setText('200')

        self.ui_dpiWidthEdit.setText('350')
        self.ui_dpiHeightEdit.setText('350')

        # Connect signal and slot
        self.ui_openButton.released.connect(self.onOpenButtonReleased)

        self.ui_qualitySlider.valueChanged.connect(self.onQualitySliderValueChanged)
        self.ui_qualityEdit.textChanged.connect(self.onQualityEditTextChanged)

        self.ui_buttonBox.accepted.connect(self.onAccepted)
        self.ui_buttonBox.rejected.connect(self.onRejected)


    @Slot()
    def onOpenButtonReleased(self):
        try:
            fpath = QFileDialog.getSaveFileName(self,
                                                self.tr("Save an image"),
                                                "",
                                                self.tr("Image files (*.png *.jpeg)"))

            fpath = fpath[0]
            if fpath:
                self.ui_filePathEdit.setText(fpath)
                fname, fext = extract_name_and_ext(fpath)
                if fext.casefold() in ['jpeg']:
                    self.ui_transparencyCheck.setEnabled(False)
                else:
                    self.ui_transparencyCheck.setEnabled(True)


            # end of if
        except Exception as err:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setWindowTitle("Open a network file")
            msg.setText("An invalid file type for image.")
            msg.exec()
        # end of except

    @Slot()
    def onAccepted(self):
        self._fpath = self.ui_filePathEdit.text()
        self._isTransparent = self.ui_transparencyCheck.isChecked()

        try:
            self._quality = int(self.ui_qualitySlider.value())
        except ValueError:
            self._quality = 100
            self.ui_qualitySlider.setValue(100)
            self.ui_qualityEdit.setText('100')

        try:
            self._scaleWidth = int(self.ui_scaleWidthEdit.text())
        except ValueError:
            self._scaleWidth = 100

        try:
            self._scaleHeight = int(self.ui_scaleWidthEdit.text())
        except ValueError:
            self._scaleHeight = 100


        try:
            self._dpiWidth = int(self.ui_dpiWidthEdit.text())
        except ValueError:
            self._dpiWidth = 350

        try:
            self._dpiHeight = int(self.ui_dpiHeightEdit.text())
        except ValueError:
            self._dpiHeight = 350

        super().accept()

    @Slot()
    def onQualitySliderValueChanged(self):
        self.ui_qualityEdit.setText(str(self.ui_qualitySlider.value()))

    @Slot()
    def onQualityEditTextChanged(self):
        if self.ui_qualityEdit.text():
            self.ui_qualitySlider.setValue(int(self.ui_qualityEdit.text()))
        else:
            self.ui_qualitySlider.setValue(0)

    @Slot()
    def onRejected(self):
        self.close()
        super().reject()


    # Read-only properties
    @property
    def fpath(self):
        return self._fpath

    @property
    def isTransparent(self):
        return self._isTransparent

    @property
    def quality(self):
        return self._quality

    @property
    def scaleWidth(self):
        return self._scaleWidth

    @property
    def scaleHeight(self):
        return self._scaleHeight

    @property
    def dpiWidth(self):
        return self._dpiWidth

    @property
    def dpiHeight(self):
        return self._dpiHeight



# end of class


from nezzle.graphics.arrows.basearrow import Triangle
from nezzle.graphics.arrows.basearrow import Hammer


class ArrowClassFactory(object):
    @staticmethod
    def create(arrow_type):
        if arrow_type.upper() == "TRIANGLE":
            return Triangle
        elif arrow_type.upper() == "HAMMER":
            return Hammer
        else:
            raise TypeError("Undefined link type: %s" % (arrow_type))

    @staticmethod
    def get_available_heads():
        return ["Triangle", "Hammer"]


import numpy as np
from qtpy.QtWidgets import QGraphicsItem
from qtpy.QtCore import QRectF
from qtpy.QtCore import QPointF
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor
from qtpy.QtGui import QPainterPath

from nezzle.graphics.baseitem import GeometryChangeItem
from nezzle.utils import length


class BaseControlPoint(GeometryChangeItem):
    def __init__(self, parent, pos, iden=None, radius=5, sticky_radius=None):
        if not iden:
            iden = parent.iden + "_CP"
        self._iden = iden
        super().__init__(iden=iden, parent=parent)

        self._radius = radius

        if not sticky_radius:
            sticky_radius = 4 * self.parent.width

        self._sticky_radius = sticky_radius

        self.rect = QRectF(-self._radius, -self._radius,
                           2*self._radius, 2*self._radius)
        self.path = QPainterPath()
        self.path.addEllipse(self.rect)

        radius_sf = 0.7*self._radius
        self.small_rect = QRectF(-radius_sf, -radius_sf,
                                 2 * radius_sf, 2 * radius_sf)

        self.setPos(pos)
        self._attr["_OLD_POS"] = QPointF(pos)

        self.setFlags(QGraphicsItem.ItemIsSelectable
                      | QGraphicsItem.ItemIsMovable
                      | QGraphicsItem.ItemIsFocusable
                      | QGraphicsItem.ItemSendsGeometryChanges)

        self.update()

    def __str__(self):
        return "ControlPoint of %s" % (str(self.parentItem()))

    @property
    def iden(self):
        return self._iden

    @property
    def radius(self):
        return self._radius

    @property
    def sticky_radius(self):
        return self._sticky_radius

    @property
    def parent(self):
        return self.parentItem()

    def boundingRect(self):
        rect = QRectF()
        if not self.parent.isSelected():
            return rect

        rect = rect.united(self.rect)
        return rect

    def shape(self):
        path = QPainterPath()
        if not self.parent.isSelected():
            return path

        path.addPath(self.path)

        return path

    def paint(self, painter, option, widget):

        if not self.parent.isSelected():
            return

        if self.isSelected():
            painter.setPen(QColor(255, 0, 102, 200))
            painter.drawEllipse(self.rect)
            painter.setBrush(QColor(255, 0, 102, 200))
            painter.drawEllipse(self.small_rect)
        else:
            painter.setPen(QColor(100, 100, 100))
            painter.drawEllipse(self.rect)
            painter.setBrush(Qt.white)
            painter.drawEllipse(self.small_rect)

        scene = self.scene()
        if scene:
            scene.invalidate()

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self.parent.setSelected(True)


class ControlPoint(BaseControlPoint):

    def is_movable(self):  # inherited from Movable
        return True

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            if self.parent.is_node_selected():
                return super().itemChange(change, self.pos())

        elif change == QGraphicsItem.ItemPositionHasChanged:
            pos = value
            self.parent['CP_POS_X'] = pos.x()
            self.parent['CP_POS_Y'] = pos.y()
            self.parent.update()

        return super().itemChange(change, value)

    def mouseReleaseEvent(self, event):
        if self.parent.is_straight() or length(self.pos()) < self.sticky_radius:
           self.setPos(QPointF(0, 0))

        return super().mouseReleaseEvent(event)


class ConnectorControlPoint(BaseControlPoint):

    def __init__(self, connectors=None, *args, **kwargs):
        self._connectors = []
        if connectors:
            for obj in connectors:
                self.append_connector(obj)

        super().__init__(*args, **kwargs)

    def is_movable(self):  # inherited from Movable
        return True

    @property
    def connectors(self):
        return self._connectors

    def append_connector(self, obj):
        if len(self._connectors) == 2:
            raise ValueError("Cannot append connector. The maximum number of connectors is two.")

        if not isinstance(obj, QPointF):
            raise TypeError("Type of connector should be QPointF, not %s." % (type(obj)))

        self._connectors.append(obj)

    def itemChange(self, change, value):

        if change == QGraphicsItem.ItemPositionChange:
            if self.parent.is_node_selected():
                return super().itemChange(change, self.pos())

            if self.isSelected():
                pos = value
                self.update_pos(pos)

            return super().itemChange(change, value)

        elif change == QGraphicsItem.ItemPositionHasChanged:
            pos = value
            self.parent[f'{self._iden}_POS_X'] = pos.x()
            self.parent[f'{self._iden}_POS_Y'] = pos.y()

            if self.isSelected():
                self.parent.update()
            return

        return super().itemChange(change, value)

    def update_pos(self, pos):
        raise NotImplementedError()

    def update_pos_by_connectors(self):
        raise NotImplementedError()


class HorizontalConnectorControlPoint(ConnectorControlPoint):
    def update_pos(self, pos):
        # Limit the range of y-axis.
        mp_y = (self.connectors[0].y() + self.connectors[1].y()) / 2  # Y-axis midpoint of connectors
        pos.setY(mp_y)

        self.connectors[0].setX(pos.x())
        self.connectors[1].setX(pos.x())

    def update_pos_by_connectors(self):
        mp_y = (self.connectors[0].y() + self.connectors[1].y()) / 2  # Y-axis midpoint of connectors
        self.setY(mp_y)
        self.setX(self.connectors[0].x())
        # Never call parent.update() here

    def mouseReleaseEvent(self, event):
        thr = 2 * self.parent.width
        for cp in self.parent.ctrl_points:
            if cp.isSelected() or isinstance(cp, HorizontalConnectorControlPoint):
                continue

            conn1, conn2 = cp.connectors
            dist_conns = np.abs(conn1.x() - conn2.x())
            if dist_conns < thr:
                self.setX(conn1.x() + conn2.x() - self.x())
                self.parent.update()
                break

        return super().mouseReleaseEvent(event)


class VerticalConnectorControlPoint(ConnectorControlPoint):
    def update_pos(self, pos):
        # Limit the range of x-axis.
        mp_x = (self.connectors[0].x() + self.connectors[1].x()) / 2  # X-axis midpoint of connectors
        pos.setX(mp_x)

        # Control the y-coordinates of connectors.
        self.connectors[0].setY(pos.y())
        self.connectors[1].setY(pos.y())

    def update_pos_by_connectors(self):
        mp_x = (self.connectors[0].x() + self.connectors[1].x()) / 2  # X-axis midpoint of connectors
        self.setX(mp_x)
        self.setY(self.connectors[0].y())
        # Never call parent.update() here

    def mouseReleaseEvent(self, event):
        thr = 2 * self.parent.width
        for cp in self.parent.ctrl_points:
            if cp.isSelected() or isinstance(cp, VerticalConnectorControlPoint):
                continue

            conn1, conn2 = cp.connectors
            dist_conns = np.abs(conn1.y() - conn2.y())
            if dist_conns < thr:
                self.setY(conn1.y() + conn2.y() - self.y())
                self.parent.update()
                break

        return super().mouseReleaseEvent(event)



from nezzle.managers.codemanager import CodeManager
from nezzle.managers.consoletabmanager import ConsoleTabManager
from nezzle.managers.historyviewmanager import HistoryViewManager
from nezzle.managers.networktreemanager import NetworkTreeManager
from nezzle.managers.networkmodelmanager import NetworkModelManager
from nezzle.managers.singleviewmanager import SingleViewManager


# [DEPRECATED TESTS]

import os
import sys
import json
import pytest

from qtpy import QtWidgets
from qtpy.QtGui import QColor
from qtpy.QtCore import Qt
from qtpy.QtCore import QPointF
from nezzle.graphics import NodeClassFactory
from nezzle.graphics import LinkClassFactory
from nezzle.graphics import ArrowClassFactory
from nezzle.graphics import LabelClassFactory
from nezzle.graphics import Network


from nezzle import fileio

app = QtWidgets.QApplication([])

# def test_json_network():
#
#     a = NodeClass("NODE_A", "A",
#                   QColor(1, 20, 200, 10),
#                   radius=5,
#                   pos=QPointF(100, 100))
#
#     b = NodeClass("NODE_B", "B",
#                   QColor(200, 20, 1, 10),
#                   radius=7,
#                   pos=QPointF(200, 200))
#
#     link_ab = ArrowLink("LINK_AB", a, b, 2, QColor(0, 0, 0, 0), '+')
#     link_ba = HammerLink("LINK_BA", b, a, 3, QColor(Qt.red), '-')
#
#     net = Network()
#
#     net.add_node(a)
#     net.add_node(b)
#
#     net.add_link(link_ab)
#     net.add_link(link_ba)
#
#     with open("_test_json_network.json", "wt") as fout:
#         fout.write(json.dumps(net.toDict()))
#
#     with open("_test_json_network.json", "rt") as fin:
#         d = json.loads(fin.read())
#
#     print("After reading from JSON")
#     net2 = Network.fromDict(d)
#
#     assert net2.to_dict() == net.toDict()
#     os.remove("_test_json_network.json")


def test_create_network():
    # Test creating network
    net = Network('ID_NETWORK')

    # Test CircleNode
    NodeClass = NodeClassFactory.create("CIRCLE_NODE")

    node = NodeClass('ID_NODE_1', 30, pos=QPointF(100, 100))
    net.add_node(node)
    node["FILL_COLOR"] = QColor(100, 50, 150, 100)
    node['BORDER_COLOR'] = Qt.red
    node["BORDER_WIDTH"] = 2
    node['BORDER_LINE_TYPE'] = Qt.DotLine
    attr = node.to_dict()

    node2 = NodeClass.from_dict(attr)
    node2['POS_X'] = -100
    node2['POS_Y'] = -300
    node2.iden = 'ID_NODE_2'
    net.add_node(node2)

    assert node2.iden == node2['ID']

    # Test EllipseNode
    NodeClass = NodeClassFactory.create("ELLIPSE_NODE")

    node3 = NodeClass('ID_NODE_3', 30, 50, pos=QPointF(-100, 100))
    net.add_node(node3)

    node4 = NodeClass.from_dict(node3.to_dict())
    net.add_node(node4)
    assert node3.to_dict() == node4.to_dict()

    node4.iden = 'ID_NODE_4'
    node4.setX(-300)
    node4.setY(200)

    assert node4.x() == node4['POS_X']
    assert node4.y() == node4['POS_Y']

    node4['POS_X'] = 1000
    node4['POS_Y'] = 2000

    assert node4.x() == node4['POS_X']
    assert node4.y() == node4['POS_Y']

    assert node3.to_dict() != node4.to_dict()

    # Create CurvedLink
    ArrowClass = ArrowClassFactory.create("TRIANGLE")
    head = ArrowClass()
    LinkClass = LinkClassFactory.create('CURVED_LINK')
    link = LinkClass('ID_LINK_1', node, node2, head=head)
    net.add_link(link)
    attr = link.to_dict()

    link2 = LinkClass.from_dict(attr, node, node2)
    link2.iden = 'ID_LINK_2'
    net.add_link(link2)

    # Create StraightLink
    ArrowClass = ArrowClassFactory.create("HAMMER")
    head = ArrowClass()
    LinkClass = LinkClassFactory.create('STRAIGHT_LINK')
    link3 = LinkClass('ID_LINK_3', node, node2, head=head)
    net.add_link(link3)

    ArrowClass = ArrowClassFactory.create("TRIANGLE")
    head = ArrowClass()
    link4 = LinkClass.from_dict(link3.to_dict(), node, node2)
    link4.iden = 'ID_LINK_4'
    link4.head = head
    net.add_link(link4)

    # Setting the size of CircleNode
    node2.width = 100
    node2.height = 200
    assert node2.width == node2.height == 2*node2.radius

    node2.radius = 25
    assert node2.width == node2.height == 2*node2.radius

    # Test creating font
    LabelClass = LabelClassFactory.create("TEXT_LABEL")
    label = LabelClass(node, "This is labels")
    label.iden = 'LABEL_1'
    net.add_label(label)
    label['TEXT_COLOR'] = Qt.red
    label['FONT_BOLD'] = True
    label['FONT_ITALIC'] = True
    label['FONT_SIZE'] = 14
    label['FONT_FAMILY'] = 'Times New Roman'
    label2 = LabelClass.from_dict(label.to_dict(), node)
    label2.iden = 'LABEL_2'

    # Test converting networks
    dict_net = net.to_dict()
    net2 = net.from_dict(dict_net)
    dict_net2 = net2.to_dict()

    assert dict_net['ID'] == dict_net2['ID']
    assert dict_net['NAME'] == dict_net2['NAME']
    assert dict_net['NEZZLE_VERSION'] == dict_net2['NEZZLE_VERSION']
    assert dict_net['BACKGROUND_COLOR'] == dict_net2['BACKGROUND_COLOR']
    assert len(dict_net["NODES"]) == len(dict_net2["NODES"])
    assert len(dict_net["LINKS"]) == len(dict_net2["LINKS"])
    assert len(dict_net["LABELS"]) == len(dict_net2["LABELS"])

def test_json_jkwon():
    net1 = fileio.read_network("jkwon_egfr_pathway.sif")
    fileio.write_network(net1, "jkwon_egfr_pathway.json")
    net2 = fileio.read_network("jkwon_egfr_pathway.json")

    assert set(net1.nodes.keys()) == set(net2.nodes.keys())
    assert set(net1.links.keys()) == set(net2.links.keys())
    assert net1.to_dict() == net2.to_dict()


def test_json_korkut_2015():
    net1 = fileio.read_network("korkut_2015.json")
    fileio.write_network(net1, "korkut_2015_02.json")
    net2 = fileio.read_network("korkut_2015_02.json")

    assert set(net1.nodes.keys()) == set(net2.nodes.keys())
    assert set(net1.links.keys()) == set(net2.links.keys())
    assert net1.to_dict() == net2.to_dict()

if __name__ == '__main__':

    test_create_network()
    test_json_jkwon()


from qtpy.QtCore import Qt
from qtpy.QtCore import QObject

from nezzle.widgets.historyview import HistoryView


class HistoryViewManager(QObject):

    def __init__(self, mw):
        super().__init__(parent=mw)
        self.mw = mw
        self._history = None
        self._history_view = self.mw.ui_historyView

        selection_model = self.history_view.selectionModel()
        selection_model.selectionChanged.connect(self.on_selection_changed)

    @property
    def history_view(self):
        return self._history_view

    @history_view.setter
    def history_view(self, view):
        if not isinstance(view, HistoryView):
            raise TypeError("%s is not HistoryView instance"%(view))

        self._history_view = view

    @property
    def stack(self):
        return self._history.stack

    @property
    def history(self):
        return self._history

    def update_history_view(self, history):
        self._history = history
        self.history_view.setStack(history.stack)

    def on_selection_changed(self, current, previous):
        model_index = current.indexes()[0]  # Get QModelIndex from QItemSelection
        scene = self.mw.sv_manager.current_scene
        selected = scene.selectedItems()

    def clear(self):
        if self._history:
            try:
                self._history.stack.clear()
            except RuntimeError as err:
                print(id(self._history.stack))
                print(err)


import os
from traceback import format_exc

from qtpy.QtCore import Qt
from qtpy.QtCore import QSize
from qtpy.QtCore import QRectF
from qtpy.QtCore import QMimeData
from qtpy.QtCore import QBuffer
from qtpy.QtCore import QDataStream
from qtpy.QtCore import QByteArray
from qtpy.QtCore import QIODevice
from qtpy.QtCore import Signal, Slot
from qtpy.QtCore import QRectF

from qtpy.QtWidgets import QWidget
from qtpy.QtWidgets import QMessageBox
from qtpy.QtWidgets import QDialog, QFileDialog
from qtpy.QtWidgets import QApplication

from qtpy.QtGui import QKeySequence
from qtpy.QtGui import QImage
from qtpy.QtGui import QPainter
from qtpy.QtGui import QClipboard



from qtpy.QtSvg import QSvgGenerator

from nezzle.dialogs.opennetworkdialog import OpenNetworkDialog
from nezzle.dialogs.exportimagedialog import ExportImageDialog

from nezzle.graphics.screen import GraphicsScene
from nezzle.graphics.screen import GraphicsView
from nezzle import fileio
from nezzle.systemstate import get_system_state
from nezzle.constants import Lock

from nezzle.graphics.nodes.nodeconverter import NodeConverter
from nezzle.graphics.nodes.basenode import BaseNode
from nezzle.graphics.nodes.ellipsenode import EllipseNode
from nezzle.graphics.nodes.rectanglenode import RectangleNode

from nezzle.graphics.links.linkconverter import LinkConverter
from nezzle.graphics.links.baselink import BaseLink
from nezzle.graphics.links.selflooplink import SelfloopLink
from nezzle.graphics.links.straightlink import StraightLink
from nezzle.graphics.links.curvedlink import CurvedLink
from nezzle.graphics.links.elbowlink import VerticalElbowLink, HorizontalElbowLink


class MenuActionHandler(QWidget):
    """
    MenuActionHandler should be a subclass of QWidget,
    otherwise @Slot decorator does not work for the instance methods.
    """

    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.initialize_actions()
        self.openNetworkDialog = OpenNetworkDialog(parent=self.mw)
        self.exportImageDialog = ExportImageDialog(parent=self.mw)

    def initialize_actions(self):
        # File
        self.mw.ui_actionOpenNetwork.triggered.connect(
            self.process_open_network)
        self.mw.ui_actionSaveNetwork.triggered.connect(
            self.process_save_network)
        self.mw.ui_actionExportImage.triggered.connect(
            self.process_export_image)

        # Edit
        # self.mw.history_manager.actionUndo.triggered.connect(self.procees_undo)
        # self.mw.history_manager.actionRedo.triggered.connect(self.process_redo)

        self.mw.ui_actionUndo.triggered.connect(self.process_undo)
        self.mw.ui_actionRedo.triggered.connect(self.process_redo)

        self.mw.ui_actionUndo.setShortcuts(QKeySequence.Undo)
        self.mw.ui_actionRedo.setShortcuts(QKeySequence.Redo)

        self.mw.ui_actionCopy.triggered.connect(self.process_copy)
        self.mw.ui_actionCopy.setShortcut(QKeySequence('Ctrl+C'))

        self.mw.ui_actionPaste.triggered.connect(self.process_paste)
        self.mw.ui_actionPaste.setShortcut(QKeySequence('Ctrl+V'))

        # View
        self.mw.ui_actionViewNavigationDock.triggered.connect(
            self.process_view_navigation_dock
        )
        self.mw.ui_navigationDock.closeEventOccured.connect(
             self.process_view_navigation_dock
        )

        self.mw.ui_actionViewConsoleDock.triggered.connect(
            self.process_view_console_dock
        )
        self.mw.ui_consoleDock.closeEventOccured.connect(
             self.process_view_console_dock
        )

        self.mw.ui_actionViewHistoryDock.triggered.connect(
            self.process_view_history_dock
        )
        self.mw.ui_historyDock.closeEventOccured.connect(
             self.process_view_history_dock
        )

        # Select -> Lock -> Lock Nodes, Lock Links, Lock Labels
        self.mw.ui_actionLockNodes.triggered.connect(
            self.process_lock_nodes
        )
        self.mw.ui_actionLockLinks.triggered.connect(
            self.process_lock_links
        )
        self.mw.ui_actionLockLabels.triggered.connect(
            self.process_lock_labels
        )

        # Select -> Convert -> To
        self.mw.ui_actionToEllipseNode.triggered.connect(
            self.process_convert_to_ellipse_node
        )
        self.mw.ui_actionToRectangleNode.triggered.connect(
            self.process_convert_to_rectangle_node
        )
        self.mw.ui_actionToStraightLink.triggered.connect(
            self.process_convert_to_straight_link
        )
        self.mw.ui_actionToCurvedLink.triggered.connect(
            self.process_convert_to_curved_link
        )
        self.mw.ui_actionToVerticalElbowLink.triggered.connect(
            self.process_convert_to_vertical_elbow_link
        )
        self.mw.ui_actionToHorizontalElbowLink.triggered.connect(
            self.process_convert_to_horizontal_elbow_link
        )


        # Select -> Select All
        self.mw.ui_actionSelectAll.triggered.connect(
            self.process_select_all
        )
        self.mw.ui_actionSelectAll.setShortcut(QKeySequence('Ctrl+A'))


        # The following objects are locked at first.
        self.mw.ui_actionLockLinks.setChecked(True)
        self.process_lock_links(True)

        self.mw.ui_actionLockLabels.setChecked(True)
        self.process_lock_labels(True)

        # Layout -> Align
        self.mw.ui_actionAlignLeft.triggered.connect(
            self.process_align_left)
        self.mw.ui_actionAlignCenter.triggered.connect(
            self.process_align_center)
        self.mw.ui_actionAlignRight.triggered.connect(
            self.process_align_right)

        self.mw.ui_actionAlignTop.triggered.connect(
            self.process_align_top)
        self.mw.ui_actionAlignMiddle.triggered.connect(
            self.process_align_middle)
        self.mw.ui_actionAlignBottom.triggered.connect(
            self.process_align_bottom)

        self.mw.ui_actionDistributeHorizontally.triggered.connect(
            self.process_distribute_horizontally)
        self.mw.ui_actionDistributeVertically.triggered.connect(
            self.process_distribute_vertically)

        self.mw.ui_menuAlign.setEnabled(False)

    @Slot()
    def process_undo(self):
        #self.mw.history_manager.undo()
        self.mw.sv_manager.view.undo_current_scene()

    @Slot()
    def process_redo(self):
        #self.mw.history_manager.redo()
        self.mw.sv_manager.view.redo_current_scene()

    # TODO: Implement Copy & Paste QGraphicsItem to the clipboard
    @Slot()
    def process_copy(self):
        # TODO: Implement copy function for copying inside the scene.
        item = self.mw.nt_manager.current_item
        if not item:
            return

        net = item.data()
        scene = net.scene
        scene.clearSelection()
        scene.clearFocus()
        brect = scene.itemsBoundingRect()
        brect.adjust(-5, -5, +10, +10)

        #print("[COPY] boundingRect:", brect, "sceneRect:", scene.sceneRect())

        width = brect.width()
        height = brect.height()

        # Create a buffer
        data = QByteArray()
        b = QBuffer(data)
        b.open(QIODevice.WriteOnly)

        # Create SVG
        svg_gen = QSvgGenerator()
        svg_gen.setOutputDevice(b)
        svg_gen.setSize(QSize(width, height))
        svg_gen.setViewBox(QRectF(0.0, 0.0, width, height))

        painter = QPainter()
        painter.begin(svg_gen)
        painter.setBackgroundMode(Qt.TransparentMode)  # Added
        painter.setRenderHint(QPainter.HighQualityAntialiasing)
        scene.render(painter, source=brect)
        painter.end()

        mimeData = QMimeData()
        mimeData.setData("image/svg+xml", b.buffer())
        QApplication.clipboard().setMimeData(mimeData, QClipboard.Clipboard)

    @Slot()
    def process_paste(self):
        #TODO: Implement paste function
        pass

    @Slot(bool)
    def process_open_network(self):

        while True:
            self.openNetworkDialog.exec()
            choice = self.openNetworkDialog.result()
            if choice == QDialog.Rejected:
                 return

            fpath = self.openNetworkDialog.fpath

            if choice == QDialog.Rejected:
                break
            elif choice == QDialog.Accepted:
                fpath = fpath.strip()
                try:
                    net = fileio.read_network(fpath, self.openNetworkDialog.link_map)
                except Exception as err:
                    err_msg = "An error has occurred during opening the file.\n%s"
                    self.show_error("Open a network file", err_msg % format_exc())
                    continue

                network_name = self.openNetworkDialog.network_name
                if network_name:
                    net.name = network_name
                else:
                    net.name = os.path.basename(fpath)

                self.mw.sv_manager.set_current_view_scene(net.scene, net.name)
                self.mw.nt_manager.append_item(net)
                # self.mw.ct_manager.update_console_variables()

            # end of if
            break
        # end of while


    @Slot(bool)
    def process_save_network(self):
        try:
            fpath = QFileDialog.getSaveFileName(self.mw,
                                                self.tr("Save a network file"),
                                                "",
                                                self.tr("Network files (*.sif *.json)"))
            fpath = fpath[0]
            if fpath:
                net = self.mw.nt_manager.current_item.data()
                if net:
                    fileio.write_network(net, fpath)
                else:
                    err_msg = "There is no selected network."
                    self.show_error("Save a network", err_msg)
            # end of if
        except Exception as err:
            err_msg = "An error has occurred during saving the file:\n%s"
            self.show_error("Save a network file", err_msg % format_exc())

    @Slot(bool)
    def process_export_image(self):
        try:
            self.exportImageDialog.exec()

            if self.exportImageDialog.result() == QDialog.Rejected:
                return

            net = self.mw.nt_manager.current_item.data()
            if net:
                fpath = self.exportImageDialog.fpath
                is_transparent = self.exportImageDialog.isTransparent
                quality = self.exportImageDialog.quality
                scale_width = self.exportImageDialog.scaleWidth
                scale_height = self.exportImageDialog.scaleHeight
                dpi_width = self.exportImageDialog.dpiWidth
                dpi_height = self.exportImageDialog.dpiHeight
                if not fpath:
                    err_msg = "File name is not designated."
                    self.show_error("Export a network image", err_msg)

                fileio.write_image(net,
                                   fpath,
                                   is_transparent,
                                   quality,
                                   scale_width, scale_height,
                                   dpi_width, dpi_height)
        except Exception as err:
            err_msg = "An error has occurred during saving the file:\n%s"
            self.show_error("Export a network image", err_msg % format_exc())

        # end of except

    def show_error(self, title, err_msg):
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setWindowTitle(title)
        msg.setText(err_msg)
        msg.exec()

    @Slot(bool)
    def process_view_navigation_dock(self, checked):
        if checked:
            self.mw.ui_navigationDock.show()
        else:
            self.mw.ui_navigationDock.hide()
        self.mw.ui_actionViewNavigationDock.setChecked(checked)

    @Slot(bool)
    def process_view_console_dock(self, checked):
        if checked:
            self.mw.ui_consoleDock.show()
        else:
            self.mw.ui_consoleDock.hide()
        self.mw.ui_actionViewConsoleDock.setChecked(checked)

    @Slot(bool)
    def process_view_history_dock(self, checked):
        if checked:
            self.mw.ui_historyDock.show()
        else:
            self.mw.ui_historyDock.hide()

        self.mw.ui_actionViewHistoryDock.setChecked(checked)

    @Slot(bool)
    def process_lock_nodes(self, checked):
        ss = get_system_state()
        ss.set_locked(Lock.NODES, checked)

    @Slot(bool)
    def process_lock_labels(self, checked):
        ss = get_system_state()
        ss.set_locked(Lock.LABELS, checked)

    @Slot(bool)
    def process_lock_links(self, checked):
        ss = get_system_state()
        ss.set_locked(Lock.LINKS, checked)

    def _process_convert_to_node(self, nodeclass):
        net = self.mw.nt_manager.current_net
        if not net:
            return

        # scene = net.scene
        # for obj in scene.selectedItems():
        #     if not isinstance(obj, BaseNode):
        #         continue
        #
        #     if type(obj) == nodeclass:
        #         continue
        #
        #     new_node = NodeConverter.convert(obj, nodeclass)
        #     net.replace_node(obj, new_node)
        # # end of for
        scene = net.scene

        old_items = []
        new_items = []
        for old_item in scene.selectedItems():
            if not isinstance(old_item, BaseNode):
                continue

            if type(old_item) == nodeclass:
                continue

            new_item = NodeConverter.convert(old_item, nodeclass)
            old_items.append(old_item)
            new_items.append(new_item)
        # end of for
        if len(old_items) > 0:
            self.mw.hv_manager.history.on_convert_nodes(net, old_items, new_items)

    @Slot()
    def process_convert_to_ellipse_node(self):
        self._process_convert_to_node(EllipseNode)

    @Slot()
    def process_convert_to_rectangle_node(self):
        self._process_convert_to_node(RectangleNode)

    def _process_convert_to_link(self, linkclass):
        net = self.mw.nt_manager.current_net
        if not net:
            return

        scene = net.scene

        old_items = []
        new_items = []
        for old_item in scene.selectedItems():
            if not isinstance(old_item, BaseLink):
                continue

            if type(old_item) == linkclass:
                continue

            new_item = LinkConverter.convert(old_item, linkclass)
            old_items.append(old_item)
            new_items.append(new_item)
        # end of for
        if len(old_items) > 0:
            self.mw.hv_manager.history.on_convert_links(net, old_items, new_items)

    @Slot()
    def process_convert_to_straight_link(self):
        self._process_convert_to_link(StraightLink)

    @Slot()
    def process_convert_to_curved_link(self):
        self._process_convert_to_link(CurvedLink)

    @Slot()
    def process_convert_to_vertical_elbow_link(self):
        self._process_convert_to_link(VerticalElbowLink)

    @Slot()
    def process_convert_to_horizontal_elbow_link(self):
        self._process_convert_to_link(HorizontalElbowLink)

    @Slot(bool)
    def process_select_all(self):
        item = self.mw.nt_manager.current_item
        if not item:
            return

        net = item.data()
        scene = net.scene
        for graphics_item in scene.items():
            graphics_item.setSelected(True)

    @Slot()
    def process_align_left(self):
        self.mw.sv_manager.view.align_objects('left')

    @Slot()
    def process_align_center(self):
        self.mw.sv_manager.view.align_objects('center')

    @Slot()
    def process_align_right(self):
        self.mw.sv_manager.view.align_objects('right')

    @Slot()
    def process_align_top(self):
        self.mw.sv_manager.view.align_objects('top')

    @Slot()
    def process_align_middle(self):
        self.mw.sv_manager.view.align_objects('middle')

    @Slot()
    def process_align_bottom(self):
        self.mw.sv_manager.view.align_objects('bottom')

    @Slot()
    def process_distribute_horizontally(self):
        self.mw.sv_manager.view.distribute_objects('horizontal')

    @Slot()
    def process_distribute_vertically(self):
        self.mw.sv_manager.view.distribute_objects('vertical')





import os
from os.path import join
from os.path import dirname
from os.path import abspath
import sys
import importlib


def reload_modules(dirs=None, modules=None):
    """Reload modules in a directory

    Args:
        dpath:

        modules:

    """
    if not dirs:
        dirs = ["."]
    elif isinstance(dirs, str):
        dirs = [dirs]

    for dpath in dirs:
        dpath = abspath(dpath)
        if os.path.isfile(dpath):
            dpath = dirname(dpath)

        sys.path.append(dpath)
        if not modules:
            modules = []
            for fname in os.listdir(dpath):
                modname, ext = os.path.splitext(fname)
                if ext != ".py":
                    continue
                modules.append(modname)

        for modname in modules:
            mod = importlib.import_module(modname)
            importlib.reload(mod)




from nezzle.systemstate import get_system_state
from nezzle.constants import Lock


def test_system_state_lock():

    ss = get_system_state()

    # Nodes
    ss.set_locked(Lock.NODES, True)
    assert(ss.is_locked(Lock.NODES))

    ss.set_locked(Lock.NODES, False)
    assert(not ss.is_locked(Lock.NODES))

    # Links
    ss.set_locked(Lock.LINKS, True)
    assert(ss.is_locked(Lock.LINKS))

    ss.set_locked(Lock.LINKS, False)
    assert(not ss.is_locked(Lock.LINKS))

    # Labels
    ss.set_locked(Lock.LABELS, True)
    assert(ss.is_locked(Lock.LABELS))

    ss.set_locked(Lock.LABELS, False)
    assert(not ss.is_locked(Lock.LABELS))

    # Setting all locked
    ss.set_locked(Lock.NODES, True)
    ss.set_locked(Lock.LINKS, True)
    ss.set_locked(Lock.LABELS, True)

    assert(ss.is_locked(Lock.NODES))
    assert(ss.is_locked(Lock.LINKS))
    assert(ss.is_locked(Lock.LABELS))

    ss.set_locked(Lock.NODES, False)
    assert(not ss.is_locked(Lock.NODES))